from .workflow_rule import WorkflowRule
from .workflow_ruletarget import WorkflowRuleTarget


class TaskRouterConfig:

    """
        TaskRouterConfig represents the filter and default_filter
        of a workflow configuration of taskrouter
    """

    def __init__(self, rules, default_target):
        self.filters = rules
        self.default_filter = default_target

    def __repr__(self):
        return self.__dict__



from __future__ import with_statement
import unittest

from mock import Mock, patch
from nose.tools import assert_equal

from twilio.rest.resources import AuthorizedConnectApps
from twilio.rest.resources import AuthorizedConnectApp


class AuthorizedConnectAppTest(unittest.TestCase):

    def setUp(self):
        self.parent = Mock()
        self.uri = "/base"
        self.auth = ("AC123", "token")
        self.resource = AuthorizedConnectApps(self.uri, self.auth)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_get(self, mock):
        mock.return_value = Mock()
        mock.return_value.content = '{"connect_app_sid": "SID"}'

        self.resource.get("SID")
        mock.assert_called_with("GET", "/base/AuthorizedConnectApps/SID",
                                auth=self.auth, use_json_extension=True)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_list(self, mock):
        mock.return_value = Mock()
        mock.return_value.content = '{"authorized_connect_apps": []}'

        self.resource.list()
        mock.assert_called_with("GET", "/base/AuthorizedConnectApps",
                                params={}, auth=self.auth,
                                use_json_extension=True)

    def test_load(self):
        instance = AuthorizedConnectApp(Mock(), "sid")
        instance.load({
            "connect_app_sid": "SID",
            "account_sid": "AC8dfe2f2358cf421cb6134cf6f217c6a3",
            "permissions": ["get-all"],
            "connect_app_friendly_name": "foo",
            "connect_app_description": "bat",
            "connect_app_company_name": "bar",
            "connect_app_homepage_url": "http://www.google.com",
            "uri": "/2010-04-01/Accounts/",
        })

        assert_equal(instance.permissions, ["get-all"])
        assert_equal(instance.sid, "SID")
        assert_equal(instance.friendly_name, "foo")
        assert_equal(instance.description, "bat")
        assert_equal(instance.homepage_url, "http://www.google.com")
        assert_equal(instance.company_name, "bar")

    def test_delete(self):
        self.assertRaises(AttributeError, getattr, self.resource, 'delete')

    def test_create(self):
        self.assertRaises(AttributeError, getattr, self.resource, 'create')

    def test_update(self):
        self.assertRaises(AttributeError, getattr, self.resource, 'update')



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class User(NextGenInstanceResource):

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this user
        """
        return self.delete_instance()


class Users(NextGenListResource):

    name = "Users"
    instance = User

    def list(self, **kwargs):
        """
        Returns a page of :class:`User` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the User instance resource representation.

        """
        return self.get_instances(kwargs)

    def create(self, id, **kwargs):
        """
        Make a phone call to a number.

        :param str id: The identity of the user.
        :param str role: The role to assign the user.

        :return: A :class:`User` object
        """
        kwargs["id"] = id
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete a given User
        """
        return self.delete_instance(sid)



import logging
import platform
import os

from twilio.exceptions import TwilioException
from twilio.rest.resources import Connection
from twilio.rest.resources import UNSET_TIMEOUT
from twilio.rest.resources import make_request
from twilio.version import __version__ as LIBRARY_VERSION


def find_credentials(environ=None):
    """
    Look in the current environment for Twilio credentials

    :param environ: the environment to check
    """
    environment = environ or os.environ
    try:
        account = environment["TWILIO_ACCOUNT_SID"]
        token = environment["TWILIO_AUTH_TOKEN"]
        return account, token
    except KeyError:
        return None, None


def set_twilio_proxy(proxy_url, proxy_port):
    Connection.set_proxy_info(proxy_url, proxy_port)


class TwilioClient(object):
    def __init__(self, account=None, token=None, base="https://api.twilio.com",
                 version="2010-04-01",
                 timeout=UNSET_TIMEOUT):
        """
        Create a Twilio API client.
        """

        # Get account credentials
        if not account or not token:
            account, token = find_credentials()
            if not account or not token:
                raise TwilioException("""
Twilio could not find your account credentials. Pass them into the
TwilioRestClient constructor like this:

    client = TwilioRestClient(account='AC38135355602040856210245275870',
                              token='2flnf5tdp7so0lmfdu3d')

Or, add your credentials to your shell environment. From the terminal, run

    echo "export TWILIO_ACCOUNT_SID=AC3813535560204085626521" >> ~/.bashrc
    echo "export TWILIO_AUTH_TOKEN=2flnf5tdp7so0lmfdu3d7wod" >> ~/.bashrc

and be sure to replace the values for the Account SID and auth token with the
values from your Twilio Account at https://www.twilio.com/user/account.
""")
        self.base = base
        self.auth = (account, token)
        self.timeout = timeout
        self.account_uri = "{0}/{1}/Accounts/{2}".format(base,
                                                         version, account)

    def request(self, path, method=None, vars=None):
        """sends a request and gets a response from the Twilio REST API

        .. deprecated:: 3.0

        :param path: the URL (relative to the endpoint URL, after the /v1
        :param url: the HTTP method to use, defaults to POST
        :param vars: for POST or PUT, a dict of data to send

        :returns: Twilio response in XML or raises an exception on error
        :raises: a :exc:`ValueError` if the path is invalid
        :raises: a :exc:`NotImplementedError` if the method is unknown

        This method is only included for backwards compatability reasons.
        It will be removed in a future version
        """
        logging.warning(":meth:`TwilioRestClient.request` is deprecated and "
                        "will be removed in a future version")

        vars = vars or {}
        params = None
        data = None

        if not path or len(path) < 1:
            raise ValueError('Invalid path parameter')
        if method and method not in ['GET', 'POST', 'DELETE', 'PUT']:
            raise NotImplementedError(
                'HTTP %s method not implemented' % method)

        if path[0] == '/':
            uri = self.base + path
        else:
            uri = self.base + '/' + path

        if method == "GET":
            params = vars
        elif method == "POST" or method == "PUT":
            data = vars

        user_agent = "twilio-python %s (python-%s)" % (
            LIBRARY_VERSION,
            platform.python_version(),
        )

        headers = {
            "User-Agent": user_agent,
            "Accept-Charset": "utf-8",
        }

        resp = make_request(method, uri, auth=self.auth, data=data,
                            params=params, headers=headers)

        return resp.content



# -*- coding: utf-8 -*-
from datetime import datetime
import unittest

from mock import Mock, sentinel, patch, ANY
from nose.tools import assert_equal, assert_true
import pytz
from six import advance_iterator

from twilio.rest.resources.imports import json
from twilio.rest.resources import Resource, NextGenListResource, NextGenInstanceResource
from twilio.rest.resources import ListResource
from twilio.rest.resources import InstanceResource

base_uri = "https://api.twilio.com/2010-04-01"
account_sid = "AC123"
auth = (account_sid, "token")


def test_resource_init():
    r = Resource(base_uri, auth)
    uri = "%s/%s" % (base_uri, r.name)

    assert_equal(r.base_uri, base_uri)
    assert_equal(r.auth, auth)
    assert_equal(r.uri, uri)


def test_equivalence():
    p = ListResource(base_uri, auth)
    r1 = p.load_instance({"sid": "AC123"})
    r2 = p.load_instance({"sid": "AC123"})
    assert_equal(r1, r2)


class ListResourceTest(unittest.TestCase):

    def setUp(self):
        self.r = ListResource(base_uri, auth)

    def testListResourceInit(self):
        uri = "%s/%s" % (base_uri, self.r.name)
        assert_equal(self.r.uri, uri)
        assert_equal(str(self.r), "<ListResource>")

    def testKeyValueLower(self):
        assert_equal(self.r.key, self.r.name.lower())

    def testIterNoKey(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {}

        self.assertRaises(StopIteration, advance_iterator, self.r.iter())

    def testRequest(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {self.r.key: [{'sid': 'foo'}]}
        advance_iterator(self.r.iter())
        self.r.request.assert_called_with("GET", "https://api.twilio.com/2010-04-01/Resources", params={})

    def testIterOneItem(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {self.r.key: [{'sid': 'foo'}]}

        items = self.r.iter()
        advance_iterator(items)

        self.assertRaises(StopIteration, advance_iterator, items)

    def testIterNoNextPage(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {self.r.key: []}

        self.assertRaises(StopIteration, advance_iterator, self.r.iter())

    def testKeyValue(self):
        self.r.key = "Hey"
        assert_equal(self.r.key, "Hey")

    def testInstanceLoading(self):
        instance = self.r.load_instance({"sid": "foo"})

        assert_true(isinstance(instance, InstanceResource))
        assert_equal(instance.sid, "foo")

    def testListResourceCreateResponse200(self):
        """We should accept 200 OK in response to a POST creating a resource."""
        self.r.request = Mock()
        return_value = Mock()
        return_value.status_code = 200
        self.r.request.return_value = return_value, {'sid': 'foo'}
        self.r.create_instance({})
        self.r.request.assert_called_with("POST", "https://api.twilio.com/2010-04-01/Resources", data={})

    def testListResourceCreateResponse201(self):
        """We should accept 201 Created in response to a POST creating a resource."""
        self.r.request = Mock()
        return_value = Mock()
        return_value.status_code = 201
        self.r.request.return_value = return_value, {'sid': 'foo'}
        self.r.create_instance({})
        self.r.request.assert_called_with("POST", "https://api.twilio.com/2010-04-01/Resources", data={})


class NextGenListResourceTest(unittest.TestCase):

    def setUp(self):
        self.r = NextGenListResource(base_uri, auth)

    def test_list_resource_init(self):
        uri = "%s/%s" % (base_uri, self.r.name)
        assert_equal(self.r.uri, uri)

    def test_iter_no_key(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {}

        self.assertRaises(StopIteration, advance_iterator, self.r.iter())

    def test_iter_key_not_present(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {'meta': {'key': 'foobars'}}

    def test_iter_request(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {'meta': {'key': 'foos'}, 'foos': [{'sid': '123'}]}
        item = advance_iterator(self.r.iter())
        self.r.request.assert_called_with("GET", "https://api.twilio.com/2010-04-01/Resources")
        assert_equal(item.sid, '123')

    def test_iter_one_item(self):
        self.r.request = Mock()
        self.r.request.return_value = Mock(), {'meta': {'key': 'foos', 'next_page_url': None}, 'foos': [{'sid': '123'}]}

        items = self.r.iter()
        advance_iterator(items)

        self.assertRaises(StopIteration, advance_iterator, items)

    def test_instance_loading(self):
        instance = self.r.load_instance({"sid": "foo"})

        assert_true(isinstance(instance, NextGenInstanceResource))
        assert_equal(instance.sid, "foo")


class testInstanceResourceInit(unittest.TestCase):

    def setUp(self):
        self.parent = ListResource(base_uri, auth)
        self.r = InstanceResource(self.parent, "123")
        self.uri = "%s/%s" % (self.parent.uri, "123")

    def testInit(self):
        assert_equal(self.r.uri, self.uri)

    def testLoad(self):
        self.r.load({"hey": "you"})
        assert_equal(self.r.hey, "you")

    def testLoadWithUri(self):
        self.r.load({"hey": "you", "uri": "foobar"})
        assert_equal(self.r.hey, "you")
        assert_equal(self.r.uri, self.uri)

    def testLoadDateCreated(self):
        self.r.load({"date_created": "Sat, 29 Sep 2012 12:47:54 +0000",
                     "uri": "foobar"})
        try:
            assert_true(hasattr(self.r.date_created, "day"))
            assert_equal(self.r.date_created.day, 29)
        except AttributeError:
            pass

    def testLoadNullDate(self):
        self.r.load({"date_created": None, "uri": "foobar"})
        assert self.r.date_created is None

    def testLoadWithFrom(self):
        self.r.load({"from": "foo"})
        assert_equal(self.r.from_, "foo")

    def testLoadSubresources(self):
        m = Mock()
        self.r.subresources = [m]
        self.r.load_subresources()
        m.assert_called_with(self.r.uri, self.r.auth, self.r.timeout)


class NextGenInstanceResourceTest(unittest.TestCase):
    def setUp(self):
        self.parent = NextGenListResource(base_uri, auth)
        self.r = NextGenInstanceResource(self.parent, "123")

    def test_load(self):
        self.r.load({"hey": "you"})
        assert_equal(self.r.hey, "you")

    def test_iso_date_parser(self):
        self.r.load({"date_created": "2015-01-01T00:00:00Z"})
        assert_equal(
            self.r.date_created,
            datetime(2015, 1, 1, 0, 0, 0, tzinfo=pytz.utc),
        )


class testTimeoutPropagation(unittest.TestCase):
    def setUp(self):
        self.parent = ListResource(base_uri, auth, timeout=sentinel.timeout)
        self.r = InstanceResource(self.parent, "123")
        self.uri = "%s/%s" % (self.parent.uri, "123")

    @patch('twilio.rest.resources.base.make_request')
    def testPassThrough(self, mock_request):
        mock_response = Mock()
        mock_response.ok = True,
        mock_response.content = json.dumps({'key': 'value'})
        mock_request.return_value = mock_response

        assert_equal(self.r.timeout, sentinel.timeout)
        assert_equal((mock_response, {'key': 'value'}), self.r.request('GET', base_uri))

        mock_request.assert_called_once_with(
            'GET',
            base_uri + '.json',
            headers=ANY,
            timeout=sentinel.timeout,
            auth=ANY
        )



from .base import set_twilio_proxy
from .client import TwilioRestClient
from .lookups import TwilioLookupsClient
from .pricing import TwilioPricingClient
from .task_router import TwilioTaskRouterClient

_hush_pyflakes = [set_twilio_proxy, TwilioRestClient, TwilioLookupsClient,
                  TwilioPricingClient, TwilioTaskRouterClient]



import time
from .. import jwt

import warnings
warnings.simplefilter('always', DeprecationWarning)

TASK_ROUTER_BASE_URL = 'https://taskrouter.twilio.com'
TASK_ROUTER_BASE_EVENTS_URL = 'https://event-bridge.twilio.com/v1/wschannels'
TASK_ROUTER_VERSION = "v1"

REQUIRED = {'required': True}
OPTIONAL = {'required': False}


def deprecated(func):
    def log_warning(*args, **kwargs):
        # stacklevel = 2 makes the warning refer to the caller of the
        # deprecation rather than the source of deprecation itself
        warnings.warn("Call to deprecated function {0}.".
                      format(func.__name__),
                      stacklevel=2,
                      category=DeprecationWarning)
        return func(*args, **kwargs)
    return log_warning


class TaskRouterCapability(object):
    def __init__(self, account_sid, auth_token, workspace_sid, channel_id):
        self.account_sid = account_sid
        self.auth_token = auth_token
        self.policies = []

        self.workspace_sid = workspace_sid
        self.channel_id = channel_id
        self.base_url = "{0}/{1}/Workspaces/{2}".format(TASK_ROUTER_BASE_URL,
                                                        TASK_ROUTER_VERSION,
                                                        workspace_sid)

        # validate the JWT
        self.validate_jwt()

        # set up resources
        self.setup_resource()

        # add permissions to GET and POST to the event-bridge channel
        self.allow_web_sockets(channel_id)

        # add permissions to fetch the instance resource
        self.add_policy(self.resource_url, "GET", True)

    @property
    def channel_prefix(self):
        return self.channel_id[0:2]

    def setup_resource(self):
        if self.channel_prefix == "WS":
            self.resource_url = self.base_url
        elif self.channel_prefix == "WK":
            self.resource_url = self.base_url + "/Workers/" + self.channel_id

            activity_url = self.base_url + "/Activities"
            self.allow(activity_url, "GET")

            reservations_url = self.base_url + "/Tasks/**"
            self.allow(reservations_url, "GET")

        elif self.channel_prefix == "WQ":
            self.resource_url = "{0}/TaskQueues/{1}".format(
                self.base_url, self.channel_id)

    def allow_web_sockets(self, channel_id):
        web_socket_url = "{0}/{1}/{2}".format(TASK_ROUTER_BASE_EVENTS_URL,
                                              self.account_sid,
                                              self.channel_id)

        self.policies.append(self.make_policy(web_socket_url, "GET", True))
        self.policies.append(self.make_policy(web_socket_url, "POST", True))

    def validate_jwt(self):
        if self.account_sid is None or self.account_sid[0:2] != "AC":
            raise ValueError('Invalid AccountSid provided: ' +
                             self.account_sid)
        if self.workspace_sid is None or self.workspace_sid[0:2] != "WS":
            raise ValueError('Invalid WorkspaceSid provided: ' +
                             self.workspace_sid)
        if self.channel_id is None:
            raise ValueError('ChannelId not provided')

        if self.channel_prefix != "WS" and self.channel_prefix != "WK" \
                and self.channel_prefix != "WQ":
            raise ValueError('Invalid ChannelId provided: ' + self.channel_id)

    def allow_fetch_subresources(self):
        self.allow(self.resource_url + "/**", "GET")

    def allow_updates(self):
        self.allow(self.resource_url, "POST")

    def allow_updates_subresources(self):
        self.allow(self.resource_url + "/**", "POST")

    def allow_delete(self):
        self.allow(self.resource_url, "DELETE")

    def allow_delete_subresources(self):
        self.allow(self.resource_url + "/**", "DELETE")

    @deprecated
    def allow_worker_fetch_attributes(self):
        if self.channel_prefix != "WK":
            raise ValueError("Deprecated func not applicable to non Worker")
        else:
            self.policies.append(self.make_policy(
                self.resource_url,
                'GET'))

    @deprecated
    def allow_worker_activity_updates(self):
        if self.channel_prefix == "WK":
            self.policies.append(self.make_policy(
                self.resource_url,
                'POST',
                True,
                post_filter={'ActivitySid': REQUIRED}))
        else:
            raise ValueError("Deprecated func not applicable to non Worker")

    @deprecated
    def allow_task_reservation_updates(self):
        if self.channel_prefix == "WK":
            tasks_url = self.base_url + "/Tasks/**"
            self.policies.append(self.make_policy(
                tasks_url,
                'POST',
                True))
        else:
            raise ValueError("Deprecated func not applicable to non Worker")

    def add_policy(self, url, method,
                   allowed, query_filter=None, post_filter=None):

        policy = self.make_policy(url, method,
                                  allowed, query_filter, post_filter)
        self.policies.append(policy)

    def allow(self, url, method, query_filter=None, post_filter=None):
        self.add_policy(url, method, True, query_filter, post_filter)

    def deny(self, url, method, query_filter=None, post_filter=None):
        self.add_policy(url, method, False, query_filter, post_filter)

    def make_policy(self, url, method,
                    allowed=True, query_filter=None, post_filter=None):

        """Create a policy dictionary for the given resource and method.
        :param str url: the resource URL to grant or deny access to
        :param str method: the HTTP method to allow or deny
        :param allowed bool: whether this request is allowed
        :param dict query_filter: specific GET parameter names
            to require or allow
        :param dict post_filter: POST parameter names
            to require or allow
        """

        return {
            'url': url,
            'method': method,
            'allow': allowed,
            'query_filter': query_filter or {},
            'post_filter': post_filter or {}
        }

    def get_resource_url(self):
        return self.resource_url

    def generate_token(self, ttl=3600):
        task_router_attributes = {
            'account_sid': self.account_sid,
            'workspace_sid': self.workspace_sid,
            'channel': self.channel_id
        }

        if self.channel_prefix == "WK":
            task_router_attributes["worker_sid"] = self.channel_id
        elif self.channel_prefix == "WQ":
            task_router_attributes["taskqueue_sid"] = self.channel_id

        return self._generate_token(ttl, task_router_attributes)

    def _generate_token(self, ttl, attributes=None):
        payload = {
            'iss': self.account_sid,
            'exp': int(time.time()) + ttl,
            'version': TASK_ROUTER_VERSION,
            'friendly_name': self.channel_id,
            'policies': self.policies,
        }

        if attributes is not None:
            payload.update(attributes)

        return jwt.encode(payload, self.auth_token, 'HS256')


class TaskRouterWorkerCapability(TaskRouterCapability):
    def __init__(self, account_sid, auth_token, workspace_sid, worker_sid):
        super(TaskRouterWorkerCapability, self).__init__(account_sid,
                                                         auth_token,
                                                         workspace_sid,
                                                         worker_sid)

        self.reservations_url = self.base_url + "/Tasks/**"
        self.activity_url = self.base_url + "/Activities"

        # add permissions to fetch the list of activities and
        # list of worker reservations
        self.allow(self.reservations_url, "GET")
        self.allow(self.activity_url, "GET")

    def setup_resource(self):
        self.resource_url = self.base_url + "/Workers/" + self.channel_id

    def allow_activity_updates(self):
        self.policies.append(self.make_policy(
            self.resource_url,
            'POST',
            True,
            post_filter={'ActivitySid': REQUIRED}))

    def allow_reservation_updates(self):
        self.policies.append(self.make_policy(
            self.reservations_url,
            'POST',
            True))


class TaskRouterTaskQueueCapability(TaskRouterCapability):
    def setup_resource(self):
        self.resource_url = self.base_url + "/TaskQueues/" + self.channel_id


class TaskRouterWorkspaceCapability(TaskRouterCapability):
    def __init__(self, account_sid, auth_token, workspace_sid):
        super(TaskRouterWorkspaceCapability, self).__init__(account_sid,
                                                            auth_token,
                                                            workspace_sid,
                                                            workspace_sid)

    def setup_resource(self):
        self.resource_url = self.base_url

from .taskrouter_config import (
    TaskRouterConfig
)

from .workflow_config import (
    WorkflowConfig
)

from .workflow_ruletarget import (
    WorkflowRuleTarget
)
from .workflow_rule import (
    WorkflowRule
)



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Role(NextGenInstanceResource):

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this role
        """
        return self.delete_instance()


class Roles(NextGenListResource):

    name = "Roles"
    instance = Role

    def list(self, **kwargs):
        """
        Returns a page of :class:`Role` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Role instance resource representation.

        """
        return self.get_instances(kwargs)

    def delete(self, sid):
        """
        Delete a given Role
        """
        return self.delete_instance(sid)



import time
import unittest

from nose.tools import assert_true, assert_equal

from twilio import jwt
from twilio.util import TwilioCapability


class JwtTest(unittest.TestCase):

    def assertIn(self, foo, bar, msg=None):
        """backport for 2.6"""
        return assert_true(foo in bar, msg=(msg or "%s not found in %s"
            % (foo, bar)))

    def test_no_permissions(self):
        token = TwilioCapability("AC123", "XXXXX")
        payload = token.payload()
        assert_equal(len(payload), 1)
        assert_equal(payload["scope"], '')

    def test_inbound_permissions(self):
        token = TwilioCapability("AC123", "XXXXX")
        token.allow_client_incoming("andy")
        payload = token.payload()

        eurl = "scope:client:incoming?clientName=andy"
        assert_equal(len(payload), 1)
        assert_equal(payload['scope'], eurl)

    def test_outbound_permissions(self):
        token = TwilioCapability("AC123", "XXXXX")
        token.allow_client_outgoing("AP123")
        payload = token.payload()

        eurl = "scope:client:outgoing?appSid=AP123"

        assert_equal(len(payload), 1)
        self.assertIn(eurl, payload['scope'])

    def test_outbound_permissions_params(self):
        token = TwilioCapability("AC123", "XXXXX")
        token.allow_client_outgoing("AP123", foobar=3)
        payload = token.payload()

        eurl = "scope:client:outgoing?appParams=foobar%3D3&appSid=AP123"
        assert_equal(payload["scope"], eurl)

    def test_events(self):
        token = TwilioCapability("AC123", "XXXXX")
        token.allow_event_stream()
        payload = token.payload()

        event_uri = "scope:stream:subscribe?path=%2F2010-04-01%2FEvents"
        assert_equal(payload["scope"], event_uri)

    def test_events_with_filters(self):
        token = TwilioCapability("AC123", "XXXXX")
        token.allow_event_stream(foobar="hey")
        payload = token.payload()

        event_uri = "scope:stream:subscribe?params=foobar%3Dhey&path=%2F2010-04-01%2FEvents"
        assert_equal(payload["scope"], event_uri)

    def test_decode(self):
        token = TwilioCapability("AC123", "XXXXX")
        token.allow_client_outgoing("AP123", foobar=3)
        token.allow_client_incoming("andy")
        token.allow_event_stream()

        outgoing_uri = "scope:client:outgoing?appParams=foobar%3D3&appSid=AP123&clientName=andy"
        incoming_uri = "scope:client:incoming?clientName=andy"
        event_uri = "scope:stream:subscribe?path=%2F2010-04-01%2FEvents"

        result = jwt.decode(token.generate(), "XXXXX")
        scope = result["scope"].split(" ")

        self.assertIn(outgoing_uri, scope)
        self.assertIn(incoming_uri, scope)
        self.assertIn(event_uri, scope)

    def setUp(self):
        self.payload = {"iss": "jeff", "exp": int(time.time()), "claim": "insanity"}

    def test_encode_decode(self):
        secret = 'secret'
        jwt_message = jwt.encode(self.payload, secret)
        decoded_payload = jwt.decode(jwt_message, secret)
        self.assertEqual(decoded_payload, self.payload)

    def test_bad_secret(self):
        right_secret = 'foo'
        bad_secret = 'bar'
        jwt_message = jwt.encode(self.payload, right_secret)
        self.assertRaises(jwt.DecodeError, jwt.decode, jwt_message, bad_secret)

    def test_decodes_valid_jwt(self):
        example_payload = {"hello": "world"}
        example_secret = "secret"
        example_jwt = "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJoZWxsbyI6ICJ3b3JsZCJ9.tvagLDLoaiJKxOKqpBXSEGy7SYSifZhjntgm9ctpyj8"
        decoded_payload = jwt.decode(example_jwt, example_secret)
        self.assertEqual(decoded_payload, example_payload)

    def test_allow_skip_verification(self):
        right_secret = 'foo'
        jwt_message = jwt.encode(self.payload, right_secret)
        decoded_payload = jwt.decode(jwt_message, verify=False)
        self.assertEqual(decoded_payload, self.payload)

    def test_no_secret(self):
        right_secret = 'foo'
        jwt_message = jwt.encode(self.payload, right_secret)
        self.assertRaises(jwt.DecodeError, jwt.decode, jwt_message)

    def test_invalid_crypto_alg(self):
        self.assertRaises(NotImplementedError, jwt.encode, self.payload, "secret", "HS1024")



import unittest

from nose.tools import assert_equal
from twilio.jwt import decode
from twilio.access_token import AccessToken

ACCOUNT_SID = 'AC123'
SIGNING_KEY_SID = 'SK123'


# python2.6 support
def assert_is_not_none(obj):
    assert obj is not None, '%r is None' % obj


class AccessTokenTest(unittest.TestCase):
    def _validate_claims(self, payload):
        assert_equal(SIGNING_KEY_SID, payload['iss'])
        assert_equal(ACCOUNT_SID, payload['sub'])
        assert_is_not_none(payload['nbf'])
        assert_is_not_none(payload['exp'])
        assert_equal(payload['nbf'] + 3600, payload['exp'])
        assert_is_not_none(payload['jti'])
        assert_equal('{0}-{1}'.format(payload['iss'], payload['nbf']),
                     payload['jti'])
        assert_is_not_none(payload['grants'])

    def test_empty_grants(self):
        scat = AccessToken(SIGNING_KEY_SID, ACCOUNT_SID, 'secret')
        token = str(scat)
        assert_is_not_none(token)
        payload = decode(token, 'secret')
        self._validate_claims(payload)
        assert_equal([], payload['grants'])

    def test_single_grant(self):
        scat = AccessToken(SIGNING_KEY_SID, ACCOUNT_SID, 'secret')
        scat.add_grant('https://api.twilio.com/**')
        token = str(scat)
        assert_is_not_none(token)
        payload = decode(token, 'secret')
        self._validate_claims(payload)
        assert_equal(1, len(payload['grants']))
        assert_equal('https://api.twilio.com/**', payload['grants'][0]['res'])
        assert_equal(['*'], payload['grants'][0]['act'])

    def test_endpoint_grant(self):
        scat = AccessToken(SIGNING_KEY_SID, ACCOUNT_SID, 'secret')
        scat.add_endpoint_grant('bob')
        token = str(scat)
        assert_is_not_none(token)
        payload = decode(token, 'secret')
        self._validate_claims(payload)
        assert_equal(1, len(payload['grants']))
        assert_equal('sip:bob@AC123.endpoint.twilio.com',
                     payload['grants'][0]['res'])
        assert_equal(['listen', 'invite'], payload['grants'][0]['act'])

    def test_rest_grant(self):
        scat = AccessToken(SIGNING_KEY_SID, ACCOUNT_SID, 'secret')
        scat.add_rest_grant('/Apps')
        token = str(scat)
        assert_is_not_none(token)
        payload = decode(token, 'secret')
        self._validate_claims(payload)
        assert_equal(1, len(payload['grants']))
        assert_equal('https://api.twilio.com/2010-04-01/Accounts/AC123/Apps',
                     payload['grants'][0]['res'])
        assert_equal(['*'], payload['grants'][0]['act'])

    def test_enable_nts(self):
        scat = AccessToken(SIGNING_KEY_SID, ACCOUNT_SID, 'secret')
        scat.enable_nts()
        token = str(scat)
        assert_is_not_none(token)
        payload = decode(token, 'secret')
        self._validate_claims(payload)
        assert_equal(1, len(payload['grants']))
        assert_equal('https://api.twilio.com/2010-04-01/Accounts/AC123/Tokens.json',
                     payload['grants'][0]['res'])
        assert_equal(['POST'], payload['grants'][0]['act'])



from .activities import (
    Activity,
    Activities
)

from .events import (
    Event,
    Events
)

from .reservations import (
    Reservation,
    Reservations
)

from .task_queues import (
    TaskQueue,
    TaskQueues
)

from .tasks import (
    Task,
    Tasks
)

from .workers import (
    Worker,
    Workers
)
from .workflows import (
    Workflow,
    Workflows
)

from .workspaces import (
    Workspace,
    Workspaces
)



from . import InstanceResource, ListResource
from six import iteritems


class ConnectApp(InstanceResource):
    """ An authorized connect app """
    pass


class ConnectApps(ListResource):
    """ A list of Connect App resources """

    name = "ConnectApps"
    instance = ConnectApp
    key = "connect_apps"

    def list(self, **kwargs):
        """
        Returns a page of :class:`ConnectApp` resources as a list. For paging
        informtion see :class:`ListResource`
        """
        return self.get_instances(kwargs)


class AuthorizedConnectApp(ConnectApp):
    """ An authorized connect app """

    id_key = "connect_app_sid"

    def load(self, entries):
        """ Translate certain parameters into others"""
        result = {}

        for k, v in iteritems(entries):
            k = k.replace("connect_app_", "")
            result[k] = v

        super(AuthorizedConnectApp, self).load(result)


class AuthorizedConnectApps(ConnectApps):
    """ A list of Authorized Connect App resources """

    name = "AuthorizedConnectApps"
    instance = AuthorizedConnectApp
    key = "authorized_connect_apps"



from mock import patch

from tests.tools import create_mock_json
from twilio.rest.conversations import TwilioConversationsClient


@patch("twilio.rest.resources.base.make_twilio_request")
def test_conversations(mock):
    client = TwilioConversationsClient("ACCOUNT_SID", "AUTH_TOKEN")
    resp = create_mock_json("tests/resources/conversations/conversation_instance.json")
    mock.return_value = resp
    client.conversations.get("CV4bbc4afc943cd2a5d29f0ce01c5656db")
    uri = "https://conversations.twilio.com/v1/Conversations/CV4bbc4afc943cd2a5d29f0ce01c5656db"
    mock.assert_called_with("GET", uri, auth=("ACCOUNT_SID", "AUTH_TOKEN"), use_json_extension=False)



from twilio.rest.base import TwilioClient
from twilio.rest.resources import UNSET_TIMEOUT
from twilio.rest.resources.conversations.conversations import ConversationsRoot


class TwilioConversationsClient(TwilioClient):
    """
    A client for accessing the Twilio Conversations API.

    XXX more verbiage here

    :param str account: Your Account Sid from `your dashboard
        <https://www.twilio.com/user/account>`_
    :param str token: Your Auth Token from `your dashboard
        <https://www.twilio.com/user/account>`_
    :param float timeout: The socket and read timeout for requests to Twilio
    """

    def __init__(self, account=None, token=None,
                 base="https://conversations.twilio.com", version="v1",
                 timeout=UNSET_TIMEOUT):

        super(TwilioConversationsClient, self).__init__(account, token, base,
                                                        version, timeout)

        self.version_uri = "%s/%s" % (base, version)
        self.conversations = ConversationsRoot(self.version_uri, self.auth,
                                               timeout)



"""
Test that make+request is making correct HTTP requests

Uses the awesome httpbin.org to validate responses
"""
import base64
import platform
import unittest

from httplib2 import Response
from nose.tools import assert_equal, raises
from mock import patch, Mock, ANY

import twilio
from twilio.rest.exceptions import TwilioRestException
from twilio.rest.resources.base import make_request, make_twilio_request
from twilio.rest.resources.connection import Connection
from twilio.rest.resources.connection import PROXY_TYPE_SOCKS5


get_headers = {
    "User-Agent": "twilio-python/{version} (Python {python_version})".format(
        version=twilio.__version__,
        python_version=platform.python_version(),
    ),
    "Accept-Charset": "utf-8",
    "Accept": "application/json",
}

post_headers = get_headers.copy()
post_headers["Content-Type"] = "application/x-www-form-urlencoded"


class MakeRequestTest(unittest.TestCase):

    @patch('twilio.rest.resources.base.Response')
    @patch('httplib2.Http')
    def test_get_params(self, http_mock, response_mock):
        http = Mock()
        http.request.return_value = (Mock(), Mock())
        http_mock.return_value = http
        make_request("GET", "http://httpbin.org/get", params={"hey": "you"})
        http.request.assert_called_with("http://httpbin.org/get?hey=you", "GET",
                                        body=None, headers=None)

    @patch('twilio.rest.resources.base.Response')
    @patch('httplib2.Http')
    def test_get_extra_params(self, http_mock, response_mock):
        http = Mock()
        http.request.return_value = (Mock(), Mock())
        http_mock.return_value = http
        make_request("GET", "http://httpbin.org/get?foo=bar", params={"hey": "you"})
        http.request.assert_called_with("http://httpbin.org/get?foo=bar&hey=you", "GET",
                                        body=None, headers=None)

    @patch('twilio.rest.resources.base.Response')
    @patch('httplib2.Http')
    def test_resp_uri(self, http_mock, response_mock):
        http = Mock()
        http.request.return_value = (Mock(), Mock())
        http_mock.return_value = http
        make_request("GET", "http://httpbin.org/get")
        http.request.assert_called_with("http://httpbin.org/get", "GET",
                                        body=None, headers=None)

    @patch('twilio.rest.resources.base.Response')
    @patch('httplib2.Http')
    def test_sequence_data(self, http_mock, response_mock):
        http = Mock()
        http.request.return_value = (Mock(), Mock())
        http_mock.return_value = http
        make_request(
            "POST",
            "http://httpbin.org/post",
            data={"a_list": ["here", "is", "some", "stuff"]},
        )
        http.request.assert_called_with(
            "http://httpbin.org/post",
            "POST",
            body="a_list=here&a_list=is&a_list=some&a_list=stuff",
            headers=None,
        )

    @patch('twilio.rest.resources.base.Response')
    @patch('httplib2.Http._conn_request')
    def test_make_request_basic_auth(self, mock_request, mock_response):
        response = Response({
            'status': '401',
            'WWW-Authenticate': 'Basic realm="Twilio API"'
        })
        mock_request.side_effect = [(response, Mock()), (Mock(), Mock())]
        make_request('GET', 'http://httpbin.org/get', auth=('AC123', 'AuthToken'))

        auth = "{0}:{1}".format('AC123', 'AuthToken')
        encoded_auth = auth.encode('utf-8')
        b64_auth = base64.b64encode(encoded_auth)

        mock_request.assert_called_with(
            ANY,
            '/get',
            'GET',
            None,
            {
                'accept-encoding': 'gzip, deflate',
                'authorization': 'Basic {0}'.format(b64_auth.decode('utf-8')),
                'user-agent': ANY,
            }
        )

    @patch('twilio.rest.resources.base.make_request')
    def test_make_twilio_request_headers(self, mock):
        url = "http://random/url"
        make_twilio_request("POST", url, use_json_extension=True)
        mock.assert_called_with("POST", "http://random/url.json",
                                headers=post_headers)

    @raises(TwilioRestException)
    @patch('twilio.rest.resources.base.make_request')
    def test_make_twilio_request_bad_data(self, mock):
        resp = Mock()
        resp.ok = False
        resp.return_value = "error"
        mock.return_value = resp

        url = "http://random/url"
        make_twilio_request("POST", url)
        mock.assert_called_with("POST", "http://random/url.json",
                                headers=post_headers)

    @patch('twilio.rest.resources.base.Response')
    @patch('httplib2.Http')
    def test_proxy_info(self, http_mock, resp_mock):
        http = Mock()
        http.request.return_value = (Mock(), Mock())
        http_mock.return_value = http
        Connection.set_proxy_info(
            'example.com',
            8080,
            proxy_type=PROXY_TYPE_SOCKS5,
        )
        make_request("GET", "http://httpbin.org/get")
        http_mock.assert_called_with(timeout=None, ca_certs=ANY, proxy_info=ANY)
        http.request.assert_called_with("http://httpbin.org/get", "GET",
                                        body=None, headers=None)
        proxy_info = http_mock.call_args[1]['proxy_info']
        assert_equal(proxy_info.proxy_host, 'example.com')
        assert_equal(proxy_info.proxy_port, 8080)
        assert_equal(proxy_info.proxy_type, PROXY_TYPE_SOCKS5)



import time
import unittest

from twilio import jwt
from twilio.task_router import TaskRouterWorkerCapability


class TaskRouterWorkerCapabilityTest(unittest.TestCase):
    def check_policy(self, method, url, policy):
        self.assertEqual(url, policy['url'])
        self.assertEqual(method, policy['method'])
        self.assertTrue(policy['allow'])
        self.assertEqual({}, policy['query_filter'])
        self.assertEqual({}, policy['post_filter'])

    def check_decoded(self, decoded, account_sid, workspace_sid, channel_id, channel_sid=None):
        self.assertEqual(decoded["iss"], account_sid)
        self.assertEqual(decoded["account_sid"], account_sid)
        self.assertEqual(decoded["workspace_sid"], workspace_sid)
        self.assertEqual(decoded["channel"], channel_id)
        self.assertEqual(decoded["version"], "v1")
        self.assertEqual(decoded["friendly_name"], channel_id)

        if 'worker_sid' in decoded.keys():
            self.assertEqual(decoded['worker_sid'], channel_sid)
        if 'taskqueue_sid' in decoded.keys():
            self.assertEqual(decoded['taskqueue_sid'], channel_sid)

    def setUp(self):
        self.account_sid = "AC123"
        self.auth_token = "foobar"
        self.workspace_sid = "WS456"
        self.worker_sid = "WK789"
        self.capability = TaskRouterWorkerCapability(self.account_sid, self.auth_token, self.workspace_sid, self.worker_sid)

    def test_generate_token(self):

        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        self.check_decoded(decoded, self.account_sid, self.workspace_sid, self.worker_sid, self.worker_sid)

    def test_generate_token_with_default_ttl(self):
        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        self.assertEqual(int(time.time()) + 3600, decoded["exp"])

    def test_generate_token_with_custom_ttl(self):
        ttl = 10000

        token = self.capability.generate_token(ttl)
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        self.assertEqual(int(time.time()) + 10000, decoded["exp"])

    def test_defaults(self):
        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        websocket_url = 'https://event-bridge.twilio.com/v1/wschannels/{0}/{1}'.format(self.account_sid, self.worker_sid)

        # expect 5 policies
        policies = decoded['policies']
        self.assertEqual(len(policies), 5)

        # should expect 5 policies
        for method, url, policy in [
            ('GET', websocket_url, policies[0]),
            ('POST', websocket_url, policies[1]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Workers/WK789", policies[2]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Tasks/**", policies[3]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Activities", policies[4])
        ]:
            yield self.check_policy, method, url, policy

    def test_allow_activity_updates(self):

        # allow activity updates to the worker
        self.capability.allow_activity_updates()

        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        policies = decoded['policies']
        self.assertEqual(len(policies), 6)
        policy = policies[5]

        url = "https://taskrouter.twilio.com/v1/Workspaces/{0}/Workers/{1}".format(self.workspace_sid, self.worker_sid)

        self.assertEqual(url, policy["url"])
        self.assertEqual("POST", policy["method"])
        self.assertTrue(policy["allow"])
        self.assertNotEqual(None, policy['post_filter'])
        self.assertEqual({}, policy['query_filter'])
        self.assertTrue(policy['post_filter']['ActivitySid'])

    def test_allow_reservation_updates(self):
        # allow reservation updates
        self.capability.allow_reservation_updates()

        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        policies = decoded['policies']
        self.assertEqual(len(policies), 6)

        policy = policies[5]

        url = "https://taskrouter.twilio.com/v1/Workspaces/{0}/Tasks/**".format(self.workspace_sid)

        self.check_policy('POST', url, policy)

if __name__ == "__main__":
    unittest.main()



from twilio.rest.base import TwilioClient
from twilio.rest.resources import (
    UNSET_TIMEOUT,
    Activities,
    Events,
    Reservations,
    TaskQueues,
    Tasks,
    Workers,
    Workflows,
    Workspaces,
)


class TwilioTaskRouterClient(TwilioClient):
    """
    A client for accessing the Twilio TaskRouter API

    :param str account: Your Account SID from `your dashboard
        <https://twilio.com/user/account>`_
    :param str token: Your Auth Token from `your dashboard
        <https://twilio.com/user/account>`_
    :param float timeout: The socket and read timeout for requests to Twilio
    """

    def __init__(self, account=None, token=None,
                 base="https://taskrouter.twilio.com", version="v1",
                 timeout=UNSET_TIMEOUT):
        """
        Create a Twilio REST API client.
        """
        super(TwilioTaskRouterClient, self).__init__(account, token, base,
                                                     version, timeout)
        self.base_uri = "{0}/{1}".format(base, version)
        self.workspace_uri = "{0}/Workspaces".format(self.base_uri)

        self.workspaces = Workspaces(self.base_uri, self.auth, timeout)

    def activities(self, workspace_sid):
        """
        Return a :class:`Activities` instance for the :class:`Activity`
        with the given workspace_sid
        """
        base_uri = "{0}/{1}".format(self.workspace_uri, workspace_sid)
        return Activities(base_uri, self.auth, self.timeout)

    def events(self, workspace_sid):
        """
        Return a :class:`Events` instance for the :class:`Event` with the given
        workspace_sid
        """
        base_uri = "{0}/{1}".format(self.workspace_uri, workspace_sid)
        return Events(base_uri, self.auth, self.timeout)

    def reservations(self, workspace_sid, task_sid):
        """
        Return a :class:`Reservations` instance for the :class:`Reservation`
        with the given workspace_sid ans task_sid
        """
        base_uri = "{0}/{1}/Tasks/{2}".format(self.workspace_uri,
                                              workspace_sid, task_sid)
        return Reservations(base_uri, self.auth, self.timeout)

    def worker_reservations(self, workspace_sid, worker_sid):
        """
        Return a :class:`Reservations` instance for the :class:`Reservation`
        with the given workspace_sid ans worker_sid
        """
        base_uri = "{0}/{1}/Workers/{2}".format(self.workspace_uri,
                                                workspace_sid, worker_sid)
        return Reservations(base_uri, self.auth, self.timeout)

    def task_queues(self, workspace_sid):
        """
        Return a :class:`TaskQueues` instance for the :class:`TaskQueue` with
        the given workspace_sid
        """
        base_uri = "{0}/{1}".format(self.workspace_uri, workspace_sid)
        return TaskQueues(base_uri, self.auth, self.timeout)

    def tasks(self, workspace_sid):
        """
        Return a :class:`Tasks` instance for the :class:`Task` with the given
        workspace_sid
        """
        base_uri = "{0}/{1}".format(self.workspace_uri, workspace_sid)
        return Tasks(base_uri, self.auth, self.timeout)

    def workers(self, workspace_sid):
        """
        Return a :class:`Workers` instance for the :class:`Worker` with the
        given workspace_sid
        """
        base_uri = "{0}/{1}".format(self.workspace_uri, workspace_sid)
        return Workers(base_uri, self.auth, self.timeout)

    def workflows(self, workspace_sid):
        """
        Return a :class:`Workflows` instance for the :class:`Workflow` with the
        given workspace_sid
        """
        base_uri = "{0}/{1}".format(self.workspace_uri, workspace_sid)
        return Workflows(base_uri, self.auth, self.timeout)



import unittest

from mock import Mock, patch

from tests.tools import create_mock_json
from twilio.rest.resources import Account


class AccountTest(unittest.TestCase):

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_usage_records_subresource(self, request):
        resp = create_mock_json("tests/resources/usage_records_list.json")
        request.return_value = resp

        mock = Mock()
        mock.uri = "/base"
        account = Account(mock, 'AC123')
        account.load_subresources()
        records = account.usage_records.list()
        self.assertEquals(len(records), 2)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_usage_triggers_subresource(self, request):
        resp = create_mock_json("tests/resources/usage_triggers_list.json")
        request.return_value = resp

        mock = Mock()
        mock.uri = "/base"
        account = Account(mock, 'AC123')
        account.load_subresources()
        triggers = account.usage_triggers.list()
        self.assertEquals(len(triggers), 2)



import unittest

from mock import patch, Mock

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.workers import Workers, Worker


AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
WORKER_SID = "WKaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
TIMEOUT = 30


class WorkerTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create(self, request):
        resp = create_mock_json('tests/resources/task_router/workers_instance.json')
        resp.status_code = 201
        request.return_value = resp

        workers = Workers(BASE_URI, AUTH, TIMEOUT)
        workers.create("Test Worker")
        exp_params = {
            'FriendlyName': "Test Worker"
        }

        request.assert_called_with("POST", "{0}/Workers".format(BASE_URI), data=exp_params, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_instance(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Workers/{1}".format(BASE_URI, WORKER_SID)
        list_resource = Workers(BASE_URI, AUTH, TIMEOUT)
        worker = Worker(list_resource, WORKER_SID)
        worker.delete()
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_list(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Workers/{1}".format(BASE_URI, WORKER_SID)
        list_resource = Workers(BASE_URI, AUTH, TIMEOUT)
        list_resource.delete(WORKER_SID)
        request.assert_called_with("DELETE", uri, auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/workers_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Workers/{1}".format(BASE_URI, WORKER_SID)
        list_resource = Workers(BASE_URI, AUTH, TIMEOUT)
        list_resource.get(WORKER_SID)
        request.assert_called_with("GET", uri, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/workers_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Workers".format(BASE_URI)
        list_resource = Workers(BASE_URI, AUTH, TIMEOUT)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/workers_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workers/{1}".format(BASE_URI, WORKER_SID)
        list_resource = Workers(BASE_URI, AUTH, TIMEOUT)
        worker = Worker(list_resource, WORKER_SID)
        worker.update(friendly_name='Test Worker')
        exp_params = {
            'FriendlyName': "Test Worker"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/workers_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workers/{1}".format(BASE_URI, WORKER_SID)
        list_resource = Workers(BASE_URI, AUTH, TIMEOUT)
        list_resource.update(WORKER_SID, friendly_name='Test Worker')
        exp_params = {
            'FriendlyName': "Test Worker"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)



import unittest

from mock import patch, Mock

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.workspaces import Workspaces, Workspace

BASE_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")
WORKSPACE_SID = "WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class WorkspaceTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create(self, request):
        resp = create_mock_json('tests/resources/task_router/workspaces_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workspaces".format(BASE_URI)
        list_resource = Workspaces(BASE_URI, AUTH)
        list_resource.create("Test Workspace", event_callback_uri="http://www.example.com", template='FIFO')
        exp_params = {
            'FriendlyName': "Test Workspace",
            'EventCallbackUri': "http://www.example.com",
            'Template': "FIFO"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_instance(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Workspaces/{1}".format(BASE_URI, WORKSPACE_SID)
        list_resource = Workspaces(BASE_URI, AUTH)
        workspace = Workspace(list_resource, WORKSPACE_SID)
        workspace.delete()
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_list(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Workspaces/{1}".format(BASE_URI, WORKSPACE_SID)
        list_resource = Workspaces(BASE_URI, AUTH)
        list_resource.delete(WORKSPACE_SID)
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/workspaces_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Workspaces/{1}".format(BASE_URI, WORKSPACE_SID)
        list_resource = Workspaces(BASE_URI, AUTH)
        list_resource.get(WORKSPACE_SID)
        request.assert_called_with("GET", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/workspaces_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Workspaces".format(BASE_URI)
        list_resource = Workspaces(BASE_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/workspaces_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workspaces/{1}".format(BASE_URI, WORKSPACE_SID)
        list_resource = Workspaces(BASE_URI, AUTH)
        workspace = Workspace(list_resource, WORKSPACE_SID)
        workspace.update(friendly_name='Test Workspace', event_callback_uri="http://www.example.com", template='FIFO')
        exp_params = {
            'FriendlyName': "Test Workspace",
            'EventCallbackUri': "http://www.example.com",
            'Template': "FIFO"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/workspaces_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workspaces/{1}".format(BASE_URI, WORKSPACE_SID)
        list_resource = Workspaces(BASE_URI, AUTH)
        list_resource.update(WORKSPACE_SID, friendly_name='Test Workspace', event_callback_uri="http://www.example.com",
                             template='FIFO')
        exp_params = {
            'FriendlyName': "Test Workspace",
            'EventCallbackUri': "http://www.example.com",
            'Template': "FIFO"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)



from .. import NextGenInstanceResource, NextGenListResource
from .statistics import Statistics


class Workspace(NextGenInstanceResource):
    """
    A Workspace resource.
    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/workspaces>_`
    for more information.

    .. attribute:: sid

        The unique ID of the Workspace

    .. attribute:: account_sid

        The ID of the account that owns this Workspace

    .. attribute:: friendly_name

        Human readable description of this workspace (for example "Sales Call
        Center" or "Customer Support Team")

    .. attribute:: default_activity_sid

        The ID of the default :class:`Activity` that will be used when new
        Workers are created in this Workspace.

    .. attribute:: default_activity_name

        The human readable name of the default activity. Read only.

    .. attribute:: timeout_activity_sid

        The ID of the Activity that will be assigned to a Worker when a
        :class:`Task` reservation times out without a response.

    .. attribute:: timeout_activity_name

        The human readable name of the timeout activity. Read only.

    .. attribute:: event_callback_url

        An optional URL where the Workspace will publish events. You can use
        this to gather data for reporting.

    .. attribute:: date_created

        The time the Workspace was created, given as UTC in ISO 8601 format.

    .. attribute:: date_updated

        The time the Workspace was last updated, given as UTC in ISO 8601
        format.
    """
    subresources = [
        Statistics,
    ]

    def delete(self):
        """
        Delete a workspace.
        """
        return self.parent.delete_instance(self.name)

    def update(self, **kwargs):
        """
        Update a workspace.
        """
        return self.parent.update_instance(self.name, kwargs)


class Workspaces(NextGenListResource):
    """ A list of Workspace resources """

    name = "Workspaces"
    instance = Workspace

    def create(self, friendly_name, **kwargs):
        """
        Create a Workspace.

        :param friendly_name: Human readable description of this workspace (for
            example "Customer Support" or "2014 Election Campaign").
        :param event_callback_url: If provided, the Workspace will publish
            events to this URL. You can use this to gather data for reporting.
            See Workspace Events for more information.
        :param template: One of the available template names. Will
            pre-configure this Workspace with the Workflow and Activities
            specified in the template. Currently "FIFO" is the only available
            template, which will configure Work Distribution Service with a set
            of default activities and a single queue for first-in, first-out
            distribution.
        """
        kwargs['friendly_name'] = friendly_name
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete the given workspace
        """
        return self.delete_instance(sid)

    def update(self, sid, **kwargs):
        """
        Update a :class:`Workspace` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)



from mock import patch, Mock
from twilio.rest.resources.keys import Keys, Key
from tests.tools import create_mock_json

ACCOUNT_SID = "ACaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
AUTH = (ACCOUNT_SID, "token")
BASE_URL = "https://api.twilio.com/2010-04-01/Accounts/{0}".format(ACCOUNT_SID)
KEY_SID = "SKaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"

list_resource = Keys(BASE_URL, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get_key(mock):
    resp = create_mock_json("tests/resources/keys_instance.json")
    mock.return_value = resp

    url = BASE_URL + "/Keys/{0}".format(KEY_SID)
    list_resource.get(KEY_SID)

    mock.assert_called_with("GET", url, auth=AUTH, use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_create_key(mock):
    resp = create_mock_json("tests/resources/keys_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    url = BASE_URL + "/Keys"
    list_resource.create(friendly_name="Fuzzy Lumpkins' SigningKey")
    params = {
        'FriendlyName': "Fuzzy Lumpkins' SigningKey"
    }

    mock.assert_called_with("POST", url, data=params, auth=AUTH, use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_update_key(mock):
    resp = create_mock_json("tests/resources/keys_instance.json")
    mock.return_value = resp

    url = BASE_URL + "/Keys/{0}".format(KEY_SID)
    list_resource.update(sid=KEY_SID, friendly_name="Fuzzy Lumpkins' SigningKey")
    params = {
        'FriendlyName': "Fuzzy Lumpkins' SigningKey"
    }

    mock.assert_called_with("POST", url, data=params, auth=AUTH, use_json_extension=True)


@patch("twilio.rest.resources.base.Resource.request")
def test_delete_key(mock):
    resp = Mock()
    resp.content = ""
    resp.status_code = 204
    mock.return_value = resp, {}

    key = Key(list_resource, KEY_SID)
    key.delete()

    url = BASE_URL + "/Keys/{0}".format(KEY_SID)
    mock.assert_called_with("DELETE", url)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_list_keys(mock):
    resp = create_mock_json("tests/resources/keys_list.json")
    mock.return_value = resp

    url = BASE_URL + "/Keys"
    list_resource.list()

    mock.assert_called_with("GET", url, params={}, auth=AUTH, use_json_extension=True)



from datetime import datetime
import unittest

from mock import patch, Mock
import pytz

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.activities import Activities, Activity

AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
ACTIVITY_SID = "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class ActivityTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create(self, request):
        resp = create_mock_json('tests/resources/task_router/activities_instance.json')
        resp.status_code = 201
        request.return_value = resp

        activities = Activities(BASE_URI, AUTH)
        activity = activities.create("Test Activity", True)
        self.assertTrue(activity is not None)
        self.assertEqual(activity.date_created, datetime(2014, 5, 14, 10, 50, 2, tzinfo=pytz.utc))
        exp_params = {
            'FriendlyName': "Test Activity",
            'Available': "true"
        }

        request.assert_called_with("POST", "{0}/Activities".format(BASE_URI),
                                   data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_instance(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Activities/{1}".format(BASE_URI, ACTIVITY_SID)
        list_resource = Activities(BASE_URI, AUTH)
        activity = Activity(list_resource, ACTIVITY_SID)
        activity.delete()
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_list(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Activities/{1}".format(BASE_URI, ACTIVITY_SID)
        list_resource = Activities(BASE_URI, AUTH)
        list_resource.delete(ACTIVITY_SID)
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/activities_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Activities/{1}".format(BASE_URI, ACTIVITY_SID)
        list_resource = Activities(BASE_URI, AUTH)
        list_resource.get(ACTIVITY_SID)
        request.assert_called_with("GET", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/activities_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Activities".format(BASE_URI)
        list_resource = Activities(BASE_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/activities_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Activities/{1}".format(BASE_URI, ACTIVITY_SID)
        list_resource = Activities(BASE_URI, AUTH)
        activity = Activity(list_resource, ACTIVITY_SID)
        activity.update(friendly_name='Test Activity', available=True)
        exp_params = {
            'FriendlyName': "Test Activity",
            'Available': "true"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/activities_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Activities/{1}".format(BASE_URI, ACTIVITY_SID)
        list_resource = Activities(BASE_URI, AUTH)
        list_resource.update(ACTIVITY_SID, friendly_name='Test Activity', available="true")
        exp_params = {
            'FriendlyName': "Test Activity",
            'Available': "true"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)



import re

from twilio.exceptions import TwilioException
from .util import change_dict_key, transform_params
from .util import UNSET_TIMEOUT
from . import InstanceResource, ListResource


TYPES = {"local": "Local", "tollfree": "TollFree", "mobile": "Mobile"}


class AvailablePhoneNumber(InstanceResource):
    """ An available phone number resource

   .. attribute:: friendly_name

      A nicely-formatted version of the phone number.

   .. attribute:: phone_number

      The phone number, in E.164 (i.e. "+1") format.

   .. attribute:: lata

      The LATA of this phone number.

   .. attribute:: rate_center

      The rate center of this phone number.

   .. attribute:: latitude

      The latitude coordinate of this phone number.

   .. attribute:: longitude

      The longitude coordinate of this phone number.

   .. attribute:: region

      The two-letter state or province abbreviation of this phone number.

   .. attribute:: postal_code

      The postal (zip) code of this phone number.

   .. attribute:: iso_country

        The country for this number

   .. attribute:: address_requirements

        Whether the phone number requires you or your customer to have an
        Address registered with Twilio. Possible values are 'none', 'any',
        'local', or 'foreign'.

   .. attribute:: beta

        (boolean) Phone numbers new to the Twilio platform are marked as beta.

    """

    def __init__(self, parent):
        # Available Phone Numbers have no sid.
        super(AvailablePhoneNumber, self).__init__(parent, "")
        self.name = ""

    def purchase(self, **kwargs):
        return self.parent.purchase(phone_number=self.phone_number,
                                    **kwargs)


class AvailablePhoneNumbers(ListResource):

    name = "AvailablePhoneNumbers"
    key = "available_phone_numbers"
    instance = AvailablePhoneNumber

    def __init__(self, base_uri, auth, timeout, phone_numbers):
        super(AvailablePhoneNumbers, self).__init__(base_uri, auth, timeout)
        self.phone_numbers = phone_numbers

    def get(self, sid):
        raise TwilioException("Individual AvailablePhoneNumbers have no sid")

    def list(self, type="local", country="US", region=None, postal_code=None,
             lata=None, rate_center=None, **kwargs):
        """
        Search for phone numbers
        """
        kwargs["in_region"] = kwargs.get("in_region", region)
        kwargs["in_postal_code"] = kwargs.get("in_postal_code", postal_code)
        kwargs["in_lata"] = kwargs.get("in_lata", lata)
        kwargs["in_rate_center"] = kwargs.get("in_rate_center", rate_center)
        params = transform_params(kwargs)

        uri = "%s/%s/%s" % (self.uri, country, TYPES[type])
        resp, page = self.request("GET", uri, params=params)

        return [self.load_instance(i) for i in page[self.key]]

    def load_instance(self, data):
        instance = self.instance(self.phone_numbers)
        instance.load(data)
        instance.load_subresources()
        return instance


class PhoneNumber(InstanceResource):
    """ An IncomingPhoneNumber object

   .. attribute:: sid

      A 34 character string that uniquely identifies this resource.

   .. attribute:: date_created

      The date that this resource was created, given as GMT RFC 2822 format.

   .. attribute:: date_updated

      The date that this resource was last updated, in GMT RFC 2822 format.

   .. attribute:: friendly_name

      A human readable descriptive text for this resource, up to 64 characters
      long. By default, the FriendlyName is a nicely formatted version of
      the phone number.

   .. attribute:: account_sid

      The unique id of the Account responsible for this phone number.

   .. attribute:: phone_number

      The incoming phone number. e.g., +16175551212 (E.164 format)

   .. attribute:: api_version

      Calls to this phone number will start a new TwiML session with this
      API version.

   .. attribute:: voice_caller_id_lookup

      Look up the caller's caller-ID name from the CNAM database (additional
      charges apply). Either true or false.

   .. attribute:: voice_url

      The URL Twilio will request when this phone number receives a call.

   .. attribute:: voice_method

      The HTTP method Twilio will use when requesting the above Url.
      Either GET or POST.

   .. attribute:: voice_fallback_url

      The URL that Twilio will request if an error occurs retrieving or
      executing the TwiML requested by Url.

   .. attribute:: voice_fallback_method

      The HTTP method Twilio will use when requesting the VoiceFallbackUrl.
      Either GET or POST.

   .. attribute:: status_callback

      The URL that Twilio will request to pass status parameters (such as
      call ended) to your application.

   .. attribute:: status_callback_method

      The HTTP method Twilio will use to make requests to the
      StatusCallback URL. Either GET or POST.

   .. attribute:: sms_url

      The URL Twilio will request when receiving an incoming SMS message
      to this number.

   .. attribute:: sms_method

      The HTTP method Twilio will use when making requests to the SmsUrl.
      Either GET or POST.

   .. attribute:: sms_fallback_url

      The URL that Twilio will request if an error occurs retrieving or
      executing the TwiML from SmsUrl.

   .. attribute:: sms_fallback_method

      The HTTP method Twilio will use when requesting the above URL.
      Either GET or POST.

   .. attribute:: uri

      The URI for this resource, relative to https://api.twilio.com.

   .. attribute:: beta

      (boolean) Phone numbers new to the Twilio platform are marked as beta.
    """

    def load(self, entries):
        """ Set the proper Account owner of this phone number """

        # Only check if entries has a uri
        if "account_sid" in entries:
            # Parse the parent's uri to get the scheme and base
            uri = re.sub(r'AC(.*)', entries["account_sid"],
                         self.parent.base_uri)

            self.parent = PhoneNumbers(
                uri,
                self.parent.auth,
                self.parent.timeout
            )
            self.base_uri = self.parent.uri

        super(PhoneNumber, self).load(entries)

    def transfer(self, account_sid):
        """
        Transfer the phone number with sid from the current account to another
        identified by account_sid
        """
        a = self.parent.transfer(self.name, account_sid)
        self.load(a.__dict__)

    def update(self, **kwargs):
        """
        Update this phone number instance.
        """
        kwargs_copy = dict(kwargs)
        change_dict_key(kwargs_copy, from_key="status_callback_url",
                        to_key="status_callback")

        a = self.parent.update(self.name, **kwargs_copy)
        self.load(a.__dict__)

    def delete(self):
        """
        Release this phone number from your account. Twilio will no longer
        answer calls to this number, and you will stop being billed the monthly
        phone number fees. The phone number will eventually be recycled and
        potentially given to another customer, so use with care. If you make a
        mistake, contact us... we may be able to give you the number back.
        """
        return self.parent.delete(self.name)


class PhoneNumbers(ListResource):

    name = "IncomingPhoneNumbers"
    key = "incoming_phone_numbers"
    instance = PhoneNumber

    def __init__(self, base_uri, auth, timeout=UNSET_TIMEOUT):
        super(PhoneNumbers, self).__init__(base_uri, auth, timeout)
        self.available_phone_numbers = \
            AvailablePhoneNumbers(base_uri, auth, timeout, self)

    def delete(self, sid):
        """
        Release this phone number from your account. Twilio will no longer
        answer calls to this number, and you will stop being billed the
        monthly phone number fees. The phone number will eventually be
        recycled and potentially given to another customer, so use with care.
        If you make a mistake, contact us... we may be able to give you the
        number back.
        """
        return self.delete_instance(sid)

    def list(self, type=None, **kwargs):
        """
        :param phone_number: Show phone numbers that match this pattern.
        :param friendly_name: Show phone numbers with this friendly name
        :param type: Filter numbers by type. Available types are
            'local', 'mobile', or 'tollfree'

        You can specify partial numbers and use '*' as a wildcard.
        """

        uri = self.uri
        if type:
            uri = "%s/%s" % (self.uri, TYPES[type])

        params = transform_params(kwargs)
        resp, page = self.request("GET", uri, params=params)

        return [self.load_instance(i) for i in page[self.key]]

    def purchase(self, status_callback_url=None, **kwargs):
        """
        Attempt to purchase the specified number. The only required parameters
        are **either** phone_number or area_code

        :returns: Returns a :class:`PhoneNumber` instance on success,
                  :data:`False` on failure
        :raises: A :exc:`TypeError` if neither phone_number or area_code
        is specified.
        """
        kwargs["StatusCallback"] = kwargs.get("status_callback",
                                              status_callback_url)

        if 'phone_number' not in kwargs and 'area_code' not in kwargs:
            raise TypeError("phone_number or area_code is required")

        number_type = kwargs.pop('type', False)
        uri = self.uri
        if number_type:
            uri = "%s/%s" % (self.uri, TYPES[number_type])

        params = transform_params(kwargs)
        resp, instance = self.request('POST', uri, data=params)

        return self.load_instance(instance)

    def search(self, **kwargs):
        """
        :param type: The type of phone number to search for.
        :param str country: Only show numbers for this country (iso2)
        :param str region: When searching the US, show numbers in this state
        :param str postal_code: Only show numbers in this area code
        :param str rate_center: US only.
        :param tuple near_lat_long: Find close numbers within Distance miles.
        :param integer distance: Search radius for a Near- query in miles.
        :param boolean beta: Whether to include numbers new to the Twilio
            platform.
        """
        return self.available_phone_numbers.list(**kwargs)

    def transfer(self, sid, account_sid):
        """
        Transfer the phone number with sid from the current account to another
        identified by account_sid
        """
        return self.update(sid, account_sid=account_sid)

    def update(self, sid, **kwargs):
        """
        Update this phone number instance
        """
        kwargs_copy = dict(kwargs)
        change_dict_key(kwargs_copy, from_key="status_callback_url",
                        to_key="status_callback")

        if "application_sid" in kwargs_copy:
            for sid_type in ["voice_application_sid", "sms_application_sid"]:
                if sid_type not in kwargs_copy:
                    kwargs_copy[sid_type] = kwargs_copy["application_sid"]
            del kwargs_copy["application_sid"]
        return self.update_instance(sid, kwargs_copy)



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Alert(NextGenInstanceResource):

    def delete(self):
        """
        Delete this alert
        """
        return self.delete_instance()


class Alerts(NextGenListResource):

    name = "Alerts"
    instance = Alert

    def list(self, before=None, after=None, **kwargs):
        """
        Returns a page of :class:`Alert` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Alert instance resource representation.

        :param date after: Only list alerts logged after this datetime
        :param date before: Only list alerts logger before this datetime
        :param log_level: If 'error', only shows errors. If 'warning',
         only show warnings
        """
        kwargs["MessageDate<"] = before
        kwargs["MessageDate>"] = after
        return self.get_instances(kwargs)

    def delete(self, sid):
        """
        Delete a given Alert
        """
        return self.delete_instance(sid)



from twilio.rest.resources.base import (
    NextGenInstanceResource,
    NextGenListResource,
)


class Event(NextGenInstanceResource):
    """An Event resource representing a state change in a TaskRouter
    Workspace.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/events>_`
    for more information.

    .. attribute:: event_type

        An identifier for this event.

    .. attribute:: account_sid

        The unique ID of the Account that owns this Event.

    .. attribute:: description

        A description of the event.

    .. attribute:: resource_type

        The type of object this event is most relevant to (:class:`Task`,
        :class:`Reservation`, :class:`Worker`).

    .. attribute:: resource_sid

        The unique ID of the object this event is most relevant to.

    .. attribute:: event_date

        The time this event was sent, in UTC ISO 8601 format.

    .. attribute:: event_data

        Data about this specific Event.
    """
    pass


class Events(NextGenListResource):
    name = "Events"
    instance = Event

    def list(self, **kwargs):
        """
        Returns a page of :class:`Event` resources as a list. For paging
            information see :class:`NextGenListResource`

        :param minutes: (Optional, Default=15) Definition of the interval in
            minutes prior to now.
        :param start_date: (Optional, Default=15 minutes prior) Filter events
            by a start date.
        :param end_date: (Optional, Default=Now) Filter events by an end date.
        :param resource_sid: (Optional) Sid of the event resource.
        :param event_type: (Optional) The type of event to filter by.
        """
        return super(Events, self).list(**kwargs)

    def get_instances(self, params):
        """
        Query the list resource for a list of InstanceResources.
        Raises a :exc:`~twilio.TwilioRestException` if requesting a page of
        results that does not exist.

        :param dict params: List of URL parameters to be included in request
        :param str page_token: Token of the page of results to retrieve
        :param int page_size: The number of results to be returned.
        :returns: -- the list of resources
        """
        return super(Events, self).get_instances(params)



from twilio.rest.resources import (
    NextGenInstanceResource,
    NextGenListResource,
    transform_params,
)


class PhoneNumber(NextGenInstanceResource):
    """
    Represents information about a phone number.

    .. attribute:: phone_number

        The phone number in normalized E.164 format, e.g. "+14158675309"

    .. attribute:: national_format

        The phone number in localized format, e.g. "(415) 867-5309"

    .. attribute:: country_code

        The ISO 3166-1 two-letter code for this phone number's country, e.g.
        "US" for United States

    .. attribute:: carrier

        A dictionary of information about the carrier responsible for this
        number, if requested.

        Contains the following:
            - mobile_country_code: the country code of the mobile carrier.
            Only populated if the number is a mobile number.
            - mobile_network_code: the network code of the mobile carrier.
            Only populated if the number is a mobile number.
            - name: the name of the carrier.
            - type: the type of the number ("mobile", "landline", or "voip")
            - error_code: the error code of the carrier info request, if one
            occurred
    """
    id_key = "phone_number"


class PhoneNumbers(NextGenListResource):
    name = "PhoneNumbers"
    instance = PhoneNumber

    def get(self, number, include_carrier_info=False, country_code=None):
        """Look up a phone number.

        :param str number: The phone number to query.
        :param bool include_carrier_info: Whether to do a carrier lookup on
            the phone number. See twilio.com for the latest pricing.
        :param str country_code: If the number is provided in a local format
        rather than E.164, specify the two-letter code of the country to parse
        the number for.
        """

        params = {}
        if country_code is not None:
            params['country_code'] = country_code

        if include_carrier_info:
            params['type'] = 'carrier'

        params = transform_params(params)
        uri = "%s/%s" % (self.uri, number)
        _, item = self.request("GET", uri, params=params)

        return self.load_instance(item)



from nose.tools import assert_equal, assert_raises
from mock import patch

from twilio.exceptions import TwilioException
from twilio.rest import TwilioRestClient
from twilio.rest.base import find_credentials


def test_creds_not_found():
    """ I shouldn't find credentials if they are not there """
    assert_equal(find_credentials({'foo': 'bar'}), (None, None))


def test_find_creds():
    """ I should find the credentials if they are there """
    env = {
        'TWILIO_ACCOUNT_SID': 'AC123',
        'foo': 'bar',
        'TWILIO_AUTH_TOKEN': '456',
    }
    assert_equal(find_credentials(env), ('AC123', '456'))


@patch("twilio.rest.base.find_credentials")
def test_creds_error(creds):
    creds.return_value = (None, None)
    assert_raises(TwilioException, TwilioRestClient)



from mock import patch
from tests.tools import create_mock_json
from twilio.rest.resources import Members

QUEUE_SID = "QU1b9faddec3d54ec18488f86c83019bf0"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")
CALL_SID = "CAaaf2e9ded94aba3e57c42a3d55be6ff2"
BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123/Queues/%s" % (
    QUEUE_SID)
TWIML_URL = "example_twiml_url"

list_resource = Members(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_members_list(mock):
    resp = create_mock_json("tests/resources/members_list.json")
    mock.return_value = resp

    uri = "%s/Members" % (BASE_URI)
    list_resource.list()

    mock.assert_called_with("GET", uri, params={}, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_members_dequeue_front(mock):
    resp = create_mock_json("tests/resources/members_instance.json")
    mock.return_value = resp

    uri = "%s/Members/Front" % (BASE_URI)
    list_resource.dequeue(TWIML_URL)

    mock.assert_called_with("POST", uri, data={"Url": TWIML_URL}, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_members_dequeue_call(mock):
    resp = create_mock_json("tests/resources/members_instance.json")
    mock.return_value = resp

    uri = "%s/Members/%s" % (BASE_URI, CALL_SID)
    list_resource.dequeue(TWIML_URL, call_sid=CALL_SID)

    mock.assert_called_with("POST", uri, data={"Url": TWIML_URL}, auth=AUTH,
                            use_json_extension=True)



from datetime import date
import unittest

from mock import Mock

from twilio.rest.resources import SmsMessages

DEFAULT = {
    'From': None,
    'DateSent<': None,
    'DateSent>': None,
    'DateSent': None,
}


class SmsTest(unittest.TestCase):

    def setUp(self):
        self.resource = SmsMessages("foo", ("sid", "token"))
        self.params = DEFAULT.copy()

    def test_list_on(self):
        self.resource.get_instances = Mock()
        self.resource.list(date_sent=date(2011, 1, 1))
        self.params['DateSent'] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_after(self):
        self.resource.get_instances = Mock()
        self.resource.list(after=date(2011, 1, 1))
        self.params['DateSent>'] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_before(self):
        self.resource.get_instances = Mock()
        self.resource.list(before=date(2011, 1, 1))
        self.params['DateSent<'] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)



from .. import NextGenInstanceResource, NextGenListResource


class Reservation(NextGenInstanceResource):
    """
    A Reservation resource.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/tasks#reservation>_`
    for more information.

    .. attribute:: sid

        The unique ID of this Reservation.

    .. attribute:: account_Sid

        The unique ID of the Account that owns this :class:`Task`.

    .. attribute:: workspace_sid

        The unique ID of the :class:`Workspace` that contains this
        :class:`Task`.

    .. attribute:: task_sid

        The unique ID of the reserved :class:`Task`.

    .. attribute:: worker_sid

        The unique ID of the reserved :class:`Worker`.

    .. attribute:: worker_name

        Human-readable description of the reserved Worker.

    .. attribute:: reservation_status

        The current status of the reservation. One of 'pending',
        'accepted', 'rejected', or 'timeout'.
    """

    def update(self, **kwargs):
        """
        Update a reservation.

        :param reservation_status: Either accepted or rejected. Specifying
            accepted means the Worker has received the Task and will process
            it. Work Distribution Service will no longer consider this task
            eligible for assignment, and no other Worker will receive this
            Task. Specifying rejected means the Worker has refused the
            assignment and Work Distribution Service will attempt to assign
            this Task to the next eligible Worker.
        :param worker_activity_sid: If rejecting a reservation, change the
            worker that is tied to this reservation to the supplied activity.
            If not provided, the WorkerPreviousActivitySid on the Reservation
            will be used.
        """
        return self.parent.update_instance(self.name, kwargs)


class Reservations(NextGenListResource):
    """ A list of Reservation resources """

    name = "Reservations"
    instance = Reservation

    def update(self, sid, **kwargs):
        """
        Update a :class:`Reservation` with the given parameters.

        :param sid: Reservation sid to update.
        :param reservation_status: Either accepted or rejected. Specifying
            accepted means the Worker has received the Task and will process
            it. Work Distribution Service will no longer consider this task
            eligible for assignment, and no other Worker will receive this
            Task. Specifying rejected means the Worker has refused the
            assignment and Work Distribution Service will attempt to assign
            this Task to the next eligible Worker.
        :param worker_activity_sid: If rejecting a reservation, change the
            worker that is tied to this reservation to the supplied activity.
            If not provided, the WorkerPreviousActivitySid on the Reservation
            will be used.
        """
        return self.update_instance(sid, kwargs)



from .taskrouter_config import TaskRouterConfig
import json


class WorkflowConfig:

    """
    WorkflowConfig represents the whole workflow config json which contains
    filters and default_filter.
    """

    def __init__(self, workflow_rules, default_target):
        # filters and default_filters
        self.task_routing = TaskRouterConfig(workflow_rules, default_target)

    def to_json(self):
        return json.dumps(self,
                          default=lambda o: o.__dict__,
                          sort_keys=True,
                          indent=4)

    @staticmethod
    def json2obj(data):
        m = json.loads(data)
        return WorkflowConfig(m['task_routing']['filters'],
                              m['task_routing']['default_filter'])



from datetime import date
from mock import patch
from nose.tools import raises, assert_equals, assert_true

from tests.tools import create_mock_json
from twilio.rest.resources import Recordings, Recording

BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")

RE_SID = "RE19e96a31ed59a5733d2c1c1c69a83a28"

recordings = Recordings(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_paging(mock):
    resp = create_mock_json("tests/resources/recordings_list.json")
    mock.return_value = resp

    uri = "%s/Recordings" % (BASE_URI)
    recordings.list(call_sid="CA123", before=date(2010, 12, 5))
    exp_params = {'CallSid': 'CA123', 'DateCreated<': '2010-12-05'}

    mock.assert_called_with("GET", uri, params=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get(mock):
    resp = create_mock_json("tests/resources/recordings_instance.json")
    mock.return_value = resp

    uri = "%s/Recordings/%s" % (BASE_URI, RE_SID)
    r = recordings.get(RE_SID)

    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=True)

    truri = "%s/Recordings/%s/Transcriptions" % (BASE_URI, RE_SID)
    assert_equals(r.transcriptions.uri, truri)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_delete_list(mock):
    resp = create_mock_json("tests/resources/recordings_instance.json")
    resp.status_code = 204
    mock.return_value = resp

    uri = "%s/Recordings/%s" % (BASE_URI, RE_SID)
    r = recordings.delete(RE_SID)

    mock.assert_called_with("DELETE", uri, auth=AUTH, use_json_extension=True)
    assert_true(r)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_delete_instance(mock):
    resp = create_mock_json("tests/resources/recordings_instance.json")
    resp.status_code = 204
    mock.return_value = resp

    uri = "%s/Recordings/%s" % (BASE_URI, RE_SID)
    rec = Recording(recordings, RE_SID)
    r = rec.delete()

    mock.assert_called_with("DELETE", uri, auth=AUTH,
                            use_json_extension=True)
    assert_true(r)


@raises(AttributeError)
def test_create():
    recordings.create


@raises(AttributeError)
def test_update():
    recordings.update



import unittest
from nose.tools import assert_raises
from twilio.rest.resources.signing_keys import SigningKeys


class SigningKeysTest(unittest.TestCase):

    def test_list(self):
        """
        Tests the Error part
        """
        keys = SigningKeys(self, [], {})
        assert_raises(AttributeError, keys.list)



from mock import patch
from tests.tools import create_mock_json
from twilio.rest import TwilioLookupsClient


@patch("twilio.rest.resources.base.make_twilio_request")
def test_phone_numbers(mock):
    client = TwilioLookupsClient("ACCOUNT_SID", "AUTH_TOKEN")
    resp = create_mock_json("tests/resources/lookups/phone_number_instance.json")
    mock.return_value = resp
    client.phone_numbers.get("+15108675309")
    uri = "https://lookups.twilio.com/v1/PhoneNumbers/+15108675309"
    mock.assert_called_with("GET", uri, params={}, auth=("ACCOUNT_SID", "AUTH_TOKEN"),
                            use_json_extension=False)



from __future__ import with_statement
import json
import unittest

from mock import Mock

from twilio.rest.resources import PhoneNumbers
from twilio.rest.resources import PhoneNumber


class PhoneNumberTest(unittest.TestCase):

    def setUp(self):
        self.parent = Mock()
        self.uri = "/base"
        self.auth = ("AC123", "token")

    def test_update_rename_status_callback_url(self):
        mock = Mock()
        mock.uri = "/base"
        instance = PhoneNumber(mock, "SID")
        instance.update(status_callback_url="http://www.example.com")
        mock.update.assert_called_with("SID", status_callback="http://www.example.com")

    def test_update_instance_rename_status_callback_url(self):
        resource = PhoneNumbers(self.uri, self.auth)
        resource.update_instance = Mock()
        resource.update("SID", status_callback_url="http://www.example.com")
        resource.update_instance.assert_called_with("SID", {"status_callback": "http://www.example.com"})

    def test_application_sid(self):
        resource = PhoneNumbers(self.uri, self.auth)
        resource.update_instance = Mock()
        resource.update("SID", application_sid="foo")
        resource.update_instance.assert_called_with(
                "SID", {"voice_application_sid": "foo", "sms_application_sid": "foo"})

    def test_voice_application_sid(self):
        resource = PhoneNumbers(self.uri, self.auth)
        resource.update_instance = Mock()
        resource.update("SID", voice_application_sid="foo")
        resource.update_instance.assert_called_with(
                "SID", {"voice_application_sid": "foo"})

    def test_sms_application_sid(self):
        resource = PhoneNumbers(self.uri, self.auth)
        resource.update_instance = Mock()
        resource.update("SID", sms_application_sid="foo")
        resource.update_instance.assert_called_with(
                "SID", {"sms_application_sid": "foo"})

    def test_status_callback_url(self):
        resource = PhoneNumbers(self.uri, self.auth)
        resource.update_instance = Mock()
        resource.update("SID", status_callback="foo")
        resource.update_instance.assert_called_with(
                "SID", {"status_callback": "foo"})

    def test_transfer(self):
        resource = PhoneNumbers(self.uri, self.auth)
        resource.update = Mock()
        resource.transfer("SID", "AC123")
        resource.update.assert_called_with("SID", account_sid="AC123")

    def test_instance_transfer(self):
        mock = Mock()
        mock.uri = "/base"
        instance = PhoneNumber(mock, "SID")
        instance.transfer("AC123")
        mock.transfer.assert_called_with("SID", "AC123")

    def test_base_uri(self):
        parent = Mock()
        parent.base_uri = ("https://api.twilio.com/2010-04-01/Accounts/"
            "AC123")
        resource = PhoneNumber(parent, "PNd2ae06cced59a5733d2c1c1c69a83a28")

        with open("tests/resources/incoming_phone_numbers_instance.json") as f:
            entry = json.load(f)
            resource.load(entry)

        self.assertEquals(resource.parent.base_uri,
            ("https://api.twilio.com/2010-04-01/Accounts/AC4bf2dafbed59a573"
             "3d2c1c1c69a83a28"))

    def test_purchase_type(self):

        types = {'local': 'Local', 'mobile': 'Mobile', 'tollfree': 'TollFree'}
        for type in ('local', 'mobile', 'tollfree'):
            resource = PhoneNumbers(self.uri, self.auth)
            resource.request = Mock()
            resource.request.return_value = (None, None)
            resource.load_instance = Mock()
            resource.purchase(type=type, phone_number='888')
            resource.request.assert_called_with('POST', self.uri + '/IncomingPhoneNumbers/' + types[type],
                                                data={'PhoneNumber': '888'})


class IncomingPhoneNumbersTest(unittest.TestCase):

    def setUp(self):
        self.resource = PhoneNumbers("http://api.twilio.com",
                                     ("user", "pass"))

    def test_mobile(self):
        request = Mock()
        request.return_value = (Mock(), {"incoming_phone_numbers": []})
        self.resource.request = request
        self.resource.list(type='mobile')

        uri = "http://api.twilio.com/IncomingPhoneNumbers/Mobile"
        request.assert_called_with("GET", uri, params={})

    def test_local(self):
        request = Mock()
        request.return_value = (Mock(), {"incoming_phone_numbers": []})
        self.resource.request = request
        self.resource.list(type='local')

        uri = "http://api.twilio.com/IncomingPhoneNumbers/Local"
        request.assert_called_with("GET", uri, params={})

    def test_toll_free(self):
        request = Mock()
        request.return_value = (Mock(), {"incoming_phone_numbers": []})
        self.resource.request = request
        self.resource.list(type='tollfree')

        uri = "http://api.twilio.com/IncomingPhoneNumbers/TollFree"
        request.assert_called_with("GET", uri, params={})



from .. import NextGenInstanceResource, NextGenListResource
from .statistics import Statistics


class Workflow(NextGenInstanceResource):
    """
    A Workflow resource.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/workflows>_`
    for more information.

    .. attribute:: sid

        The unique ID of the Workflow.

    .. attribute:: account_sid

        The ID of the account that owns this Workflow.

    .. attribute:: workspace_sid

        The ID of the :class:`Workspace` that contains this Workflow.

    .. attribute:: friendly_name

        A human-readable description of this Workflow.

    .. attribute:: assignment_callback_url

        The URL that will be called whenever a :class:`Task` managed by this
        Workflow is assigned to a :class:`Worker`.

    .. attribute:: fallback_assignment_callback_url

        If the request to the `assignment_callback_url` fails, the assignment
        callback will be made to this URL.

    .. attribute:: configuration

        A JSON document configuring the rules for this workflow.

    .. attribute:: task_reservation_timeout

        Determines how long TaskRouter will wait for a confirmation response
        from your application after assigning a Task to a worker. Defaults to
        120 seconds.

    .. attribute:: date_created

        The time this workflow was created, as UTC in ISO 8601 format.

    .. attribute:: date_updated

        The time this workflow was last updated, as UTC in ISO 8601 format.
    """
    subresources = [
        Statistics,
    ]

    def delete(self):
        """
        Delete a workflow.
        """
        return self.parent.delete_instance(self.name)

    def update(self, **kwargs):
        """
        Update a workflow.
        """
        return self.parent.update_instance(self.name, kwargs)


class Workflows(NextGenListResource):
    """ A list of Workflow resources """

    name = "Workflows"
    instance = Workflow

    def create(self, friendly_name, configuration, assignment_callback_url,
               **kwargs):
        """
        Create a Workflow.

        :param friendly_name: A string representing a human readable name for
            this Workflow. Examples include 'Inbound Call Workflow' or '2014
            Outbound Campaign'.
        :param configuration: JSON document configuring the rules for this
            Workflow.
        :param assignment_callback_url: A valid URL for the application that
            will process task assignment events.
        :param fallback_assignment_callback_url: If the request to the
            assignment_callback_url fails, the assignment callback will be made
            to this URL.
        :param task_reservation_timeout: An integer value controlling how long
            in seconds Work Distribution Service will wait for a confirmation
            response from your application after assigning a Task to a worker.
            Defaults to 120 seconds.
        """
        kwargs['friendly_name'] = friendly_name
        kwargs['configuration'] = configuration
        kwargs['assignment_callback_url'] = assignment_callback_url
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete the given workflow
        """
        return self.delete_instance(sid)

    def update(self, sid, **kwargs):
        """
        Update a :class:`Workflow` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)



from .util import transform_params
from . import InstanceResource, ListResource


class Sandbox(InstanceResource):

    id_key = "pin"

    def update(self, **kwargs):
        """
        Update your Twilio Sandbox
        """
        a = self.parent.update(**kwargs)
        self.load(a.__dict__)


class Sandboxes(ListResource):

    name = "Sandbox"
    instance = Sandbox

    def get(self):
        """Request the specified instance resource"""
        return self.get_instance(self.uri)

    def update(self, **kwargs):
        """
        Update your Twilio Sandbox
        """
        resp, entry = self.request("POST", self.uri,
                                   body=transform_params(kwargs))
        return self.create_instance(entry)



from mock import patch, Mock
from nose.tools import raises
from twilio.rest.resources import Transcriptions, Transcription
from tests.tools import create_mock_json

BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")

transcriptions = Transcriptions(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_paging(mock):
    resp = create_mock_json("tests/resources/transcriptions_list.json")
    mock.return_value = resp

    uri = "%s/Transcriptions" % (BASE_URI)
    transcriptions.list(page=2)

    mock.assert_called_with("GET", uri, params={"Page": 2}, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get(mock):
    resp = create_mock_json("tests/resources/transcriptions_instance.json")
    mock.return_value = resp

    uri = "%s/Transcriptions/TR123" % (BASE_URI)
    transcriptions.get("TR123")

    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.Resource.request")
def test_delete_transcription(req):
    """ Deleting a transcription should work """
    resp = Mock()
    resp.content = ""
    resp.status_code = 204
    req.return_value = resp, {}

    app = Transcription(transcriptions, "TR123")
    app.delete()
    uri = "https://api.twilio.com/2010-04-01/Accounts/AC123/Transcriptions/TR123"
    req.assert_called_with("DELETE", uri)


@raises(AttributeError)
def test_create():
    transcriptions.create


@raises(AttributeError)
def test_update():
    transcriptions.update



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Member(NextGenInstanceResource):

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this member
        """
        return self.delete_instance()


class Members(NextGenListResource):

    name = "Members"
    instance = Member

    def list(self, **kwargs):
        """
        Returns a page of :class:`Member` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Member instance resource representation.

        """
        return self.get_instances(kwargs)

    def create(self, identity, **kwargs):
        """
        Create a Member.

        :param str identity: The identity of the user.
        :param str role: The role to assign the member.

        :return: A :class:`Member` object
        """
        kwargs["identity"] = identity
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete a given Member
        """
        return self.delete_instance(sid)



from . import InstanceResource, ListResource
from .applications import Applications
from .notifications import Notifications
from .recordings import Transcriptions, Recordings
from .calls import Calls
from .sms_messages import Sms
from .caller_ids import CallerIds
from .phone_numbers import PhoneNumbers
from .conferences import Conferences
from .connect_apps import ConnectApps, AuthorizedConnectApps
from .queues import Queues
from .keys import Keys
from .usage import UsageRecords, UsageTriggers
from .messages import Messages
from .media import MediaList
from .sip import Sip
from .signing_keys import SigningKeys


class Account(InstanceResource):
    """ An Account resource

    """

    ACTIVE = "active"        #: A constant representing an active account
    SUSPENDED = "suspended"  #: A constant representing a suspended account
    CLOSED = "closed"        #: A constant representing a closed account

    subresources = [
        Applications,
        Notifications,
        Transcriptions,
        Recordings,
        Calls,
        Sms,
        CallerIds,
        PhoneNumbers,
        Conferences,
        ConnectApps,
        Queues,
        AuthorizedConnectApps,
        UsageRecords,
        UsageTriggers,
        MediaList,
        Messages,
        SigningKeys,
        Sip,
        Keys,
    ]

    def update(self, **kwargs):
        """
        Update the status of an account.

        :param friendly_name: Update the description of this account.
        :param status: Alter the status of this account

        Use :const:`CLOSED` to irreversibly close this account,
        :const:`SUSPENDED` to temporarily suspend it, or :const:`ACTIVE`
        to reactivate it.
        """
        self.update_instance(**kwargs)

    def close(self):
        """
        Permenently deactivate this account
        """
        return self.update_instance(status=Account.CLOSED)

    def suspend(self):
        """
        Temporarily suspend this account
        """
        return self.update_instance(status=Account.SUSPENDED)

    def activate(self):
        """
        Reactivate this account
        """
        return self.update_instance(status=Account.ACTIVE)


class Accounts(ListResource):
    """ A list of Account resources """

    name = "Accounts"
    instance = Account

    def list(self, **kwargs):
        """
        Returns a page of :class:`Account` resources as a list. For paging
        informtion see :class:`ListResource`

        :param date friendly_name: Only list accounts with this friendly name
        :param date status: Only list accounts with this status
        """
        return self.get_instances(kwargs)

    def update(self, sid, **kwargs):
        """
        :param sid: Account identifier
        :param friendly_name: Update the description of this account.
        :param status: Alter the status of this account

        Use :data:`CLOSED` to irreversibly close this account,
        :data:`SUSPENDED` to temporarily suspend it, or :data:`ACTIVE`
        to reactivate it.
        """
        return self.update_instance(sid, kwargs)

    def close(self, sid):
        """
        Permenently deactivate an account, Alias to update
        """
        return self.update(sid, status=Account.CLOSED)

    def suspend(self, sid):
        """
        Temporarily suspend an account, Alias to update
        """
        return self.update(sid, status=Account.SUSPENDED)

    def activate(self, sid):
        """
        Reactivate an account, Alias to update
        """
        return self.update(sid, status=Account.ACTIVE)

    def create(self, **kwargs):
        """
        Returns a newly created sub account resource.

        :param friendly_name: Update the description of this account.
        """
        return self.create_instance(kwargs)



import unittest
import json

from twilio.task_router.workflow_config import WorkflowConfig
from twilio.task_router.workflow_rule import WorkflowRule
from twilio.task_router.workflow_ruletarget import WorkflowRuleTarget


class WorkflowConfigTest(unittest.TestCase):
        def test_to_json(self):
            rules = [
                WorkflowRule("1==1", [WorkflowRuleTarget("WQeae4fc2f4db7f377c5d3758fb08b79b7", "1==1", 1, 20)], "SomeQ"),
                WorkflowRule("1==1", [WorkflowRuleTarget("WQ19ebe92fb33522f018b5a31d805d94da", "1==1", 1, 210)], "SomeOtherQ")
            ]
            def_target = WorkflowRuleTarget("WQ9963154bf3122d0a0558f3763951d916", "1==1", None, None)
            config = WorkflowConfig(rules, def_target)
            self.assertEqual(True, self.is_json(config.to_json()))

        def test_from_json(self):

            data = {
                       'task_routing':
                       {
                           'filters': [
                               {
                                   'expression': 'type == "sales"',
                                   'friendly_name': 'Sales',
                                   'targets': [
                                       {
                                           'queue': 'WQec62de0e1148b8477f2e24579779c8b1',
                                           'expression': 'task.language IN worker.languages'
                                       }
                                   ]
                               },
                               {
                                   'expression': 'type == "marketing"',
                                   'friendly_name': 'Marketing',
                                   'targets': [
                                       {
                                           'queue': 'WQ2acd4c1a41ffadce5d1bac9e1ce2fa9f',
                                           'expression': 'task.language IN worker.languages'
                                       }
                                   ]
                               },
                               {
                                   'expression': 'type == "support"',
                                   'friendly_name': 'Support',
                                   'targets': [
                                       {
                                           'queue': 'WQe5eb317eb23500ade45087ea6522896c',
                                           'expression': 'task.language IN worker.languages'
                                       }
                                   ]
                               }
                           ],
                           'default_filter':
                           {
                               'queue': 'WQ05f810d2d130344fd56e3c91ece2e594'
                           }
                       }
                   }

            config = WorkflowConfig.json2obj(json.dumps(data))
            self.assertEqual(3, len(config.task_routing.filters))
            self.assertEqual(1, len(config.task_routing.default_filter))

        def test_from_json2(self):

            data = {'task_routing': {'filters': [{'friendly_name': 'SomeQ', 'expression': '1==1', 'targets': [
                {'priority': 1, 'queue': 'WQXXXX', 'expression': '1==1', 'timeout': 20}]},
                                                  {'friendly_name': 'SomeOtherQ', 'expression': '1==1',
                                                   'targets': [
                                                       {'priority': 1, 'queue': 'WQXXXX', 'expression': '1==1',
                                                        'timeout': 20}]}],
                                      'default_filter': {'priority': None, 'queue': 'WQYYYYY', 'expression': None,
                                                         'timeout': None}}}
            config = WorkflowConfig.json2obj(json.dumps(data))
            self.assertEqual(2, len(config.task_routing.filters))
            self.assertEqual(4, len(config.task_routing.default_filter))

        def is_json(self, myjson):
            try:
                json.loads(myjson)
            except ValueError as e:
                print(e)
                return False
            return True



from .channels import Channels
from .roles import Roles
from .users import Users
from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Service(NextGenInstanceResource):

    subresources = [
        Channels,
        Roles,
        Users
    ]

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this service
        """
        return self.delete_instance()


class Services(NextGenListResource):

    name = "Services"
    instance = Service

    def list(self, **kwargs):
        """
        Returns a page of :class:`Service` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Service instance resource representation.

        """
        return self.get_instances(kwargs)

    def create(self, friendly_name, **kwargs):
        """
        Create a service.

        :param str friendly_name: The friendly name for the service

        :return: A :class:`Service` object
        """
        kwargs["friendly_name"] = friendly_name
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete a given Service
        """
        return self.delete_instance(sid)



from .voice import (
    Voice,
    VoiceCountry,
    VoiceCountries,
    VoiceNumber,
    VoiceNumbers,
)

from .phone_numbers import (
    PhoneNumberCountries,
    PhoneNumberCountry,
    PhoneNumbers,
)



from mock import patch
from tests.tools import create_mock_json
from twilio.rest.resources import Queues, Queue

BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")
QUEUE_SID = "QU1b9faddec3d54ec18488f86c83019bf0"
CALL_SID = "CAaaf2e9ded94aba3e57c42a3d55be6ff2"

list_resource = Queues(BASE_URI, AUTH)
instance_resource = Queue(list_resource, QUEUE_SID)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_queues_list(mock):
    resp = create_mock_json("tests/resources/queues_list.json")
    mock.return_value = resp

    uri = "%s/Queues" % (BASE_URI)
    list_resource.list()

    mock.assert_called_with("GET", uri, params={}, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_queues_create(mock):
    resp = create_mock_json("tests/resources/queues_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    uri = "%s/Queues" % (BASE_URI)
    list_resource.create('test', max_size=9001)

    mock.assert_called_with("POST", uri,
                            data={'FriendlyName': 'test', 'MaxSize': 9001},
                            auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_queues_get(mock):
    resp = create_mock_json("tests/resources/queues_instance.json")
    mock.return_value = resp

    uri = "%s/Queues/%s" % (BASE_URI, QUEUE_SID)
    list_resource.get(QUEUE_SID)
    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_queue_update(mock):
    resp = create_mock_json("tests/resources/queues_instance.json")
    mock.return_value = resp

    uri = "%s/Queues/%s" % (BASE_URI, QUEUE_SID)
    instance_resource.update(friendly_name='QQ')

    mock.assert_called_with("POST", uri,
                            data={'FriendlyName': 'QQ'}, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_queue_delete(mock):
    resp = create_mock_json("tests/resources/queues_instance.json")
    mock.return_value = resp

    uri = "%s/Queues/%s" % (BASE_URI, QUEUE_SID)
    instance_resource.delete()

    mock.assert_called_with("DELETE", uri, auth=AUTH, use_json_extension=True)



from .. import NextGenInstanceResource, NextGenListResource


class Activity(NextGenInstanceResource):
    """
    An Activity resource.
    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/activities>_`
    for more information.

    .. attribute:: sid

        The unique ID for this Activity.

    .. attribute:: account_sid

        The unique ID of the Account that owns this Activity.

    .. attribute:: workspace_sid

        The unique ID of the :class:`Workspace` that owns this Activity.

    .. attribute:: friendly_name

        A human-readable name for the Activity, such as 'on-call', 'break',
        'email', etc. These names will be used to calculate and expose
        statistics about workers, and give you visibility into the state of
        each of your workers.

    .. attribute:: available

        Boolean value indicating whether the worker should be eligible to
        receive a Task when they occupy this Activity. For example, in an
        activity called 'On Call', the worker would be unavailable to receive
        additional Task assignments.

    .. attribute:: date_created

        The date this Activity was created, given as UTC in ISO 8601 format.

    .. attribute:: date_updated

        The date this Activity was last updated, given as UTC in ISO 8601
        format.
    """

    def delete(self):
        """
        Delete an activity.
        """
        return self.parent.delete_instance(self.name)

    def update(self, **kwargs):
        """
        Update an activity.
        """
        return self.parent.update_instance(self.name, kwargs)


class Activities(NextGenListResource):
    """ A list of Activity resources """

    name = "Activities"
    instance = Activity

    def create(self, friendly_name, available):
        """
        Create an Activity.

        :param friendly_name: A human-readable name for the activity, such as
            'On Call', 'Break', 'Email', etc. Must be unique in this Workspace.
            These names will be used to calculate and expose statistics about
            workers, and give you visibility into the state of each of your
            workers.
        :param available: Boolean value indicating whether the worker should be
            eligible to receive a Task when they occupy this Activity. For
            example, a call center might have an activity named 'On Call' with
            an availability set to 'false'.
        """
        return self.create_instance({'friendly_name': friendly_name,
                                     'available': available})

    def delete(self, sid):
        """
        Delete the given activity
        """
        return self.delete_instance(sid)

    def update(self, sid, **kwargs):
        """
        Update an :class:`Activity` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)



from .. import NextGenInstanceResource, NextGenListResource
from .statistics import Statistics


class TaskQueue(NextGenInstanceResource):
    """
    A TaskQueue resource.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/taskqueues>_`
    for more information.

    .. attribute:: sid

        The unique ID of this TaskQueue.

    .. attribute:: account_sid

        The unique ID of the account that owns this TaskQueue.

    .. attribute:: workspace_sid

        The unique ID of the :class:`Workspace` that contains this TaskQueue.

    .. attribute:: friendly_name

        Human-readable description of this TaskQueue (e.g. "Customer Support"
        or "Sales").

    .. attribute:: target_workers

        The worker selection expressions associated with this TaskQueue.

    .. attribute:: reservation_activity_sid

        The :class:`Activity` to assign a :class:`Worker` when they are
        reserved for a :class:`Task` from this TaskQueue. Defaults to
        'Reserved for Task'.

    .. attribute:: assignment_activity_sid

        The Activity to assign a Worker when they accept a Task from this
        Taskqueue. Defaults to 'Unavailable for Assignment'.
    """
    subresources = [
        Statistics,
    ]

    def delete(self):
        """
        Delete a task queue.
        """
        return self.parent.delete_instance(self.name)

    def update(self, **kwargs):
        """
        Update a task queue.
        """
        return self.parent.update_instance(self.name, kwargs)


class TaskQueues(NextGenListResource):
    """ A list of TaskQueue resources """

    name = "TaskQueues"
    instance = TaskQueue
    key = "task_queues"

    def __init__(self, base_uri, auth, timeout, **kwargs):
        super(TaskQueues, self).__init__(base_uri, auth, timeout, **kwargs)
        self.statistics = Statistics(self.uri, auth, **kwargs)

    def create(self, friendly_name, assignment_activity_sid,
               reservation_activity_sid, **kwargs):
        """
        Create a TaskQueue.

        :param friendly_name: Human readable description of this TaskQueue (for
            example "Support - Tier 1", "Sales" or "Escalation")
        :param assignment_activity_sid: ActivitySID to assign workers once a
            task is assigned for them.
        :param reservation_activity_sid: ActivitySID to assign workers once a
            task is reserved for them.
        """
        kwargs['friendly_name'] = friendly_name
        kwargs['assignment_activity_sid'] = assignment_activity_sid
        kwargs['reservation_activity_sid'] = reservation_activity_sid
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete the given task queue
        """
        return self.delete_instance(sid)

    def update(self, sid, **kwargs):
        """
        Update a :class:`TaskQueue` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)



import unittest

from mock import patch, Mock

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.tasks import Tasks, Task


AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
TASK_SID = "WTaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class TaskTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create(self, request):
        resp = create_mock_json('tests/resources/task_router/tasks_instance.json')
        resp.status_code = 201
        request.return_value = resp

        tasks = Tasks(BASE_URI, AUTH)
        tasks.create("attributes", "WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", timeout=60)
        exp_params = {
            'Attributes': "attributes",
            'WorkflowSid': "WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            'Timeout': 60
        }

        request.assert_called_with("POST", "{0}/Tasks".format(BASE_URI),
                                   data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_instance(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Tasks/{1}".format(BASE_URI, TASK_SID)
        list_resource = Tasks(BASE_URI, AUTH)
        task = Task(list_resource, TASK_SID)
        task.delete()
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_list(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Tasks/{1}".format(BASE_URI, TASK_SID)
        list_resource = Tasks(BASE_URI, AUTH)
        list_resource.delete(TASK_SID)
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/tasks_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Tasks/{1}".format(BASE_URI, TASK_SID)
        list_resource = Tasks(BASE_URI, AUTH)
        list_resource.get(TASK_SID)
        request.assert_called_with("GET", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/tasks_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Tasks".format(BASE_URI)
        list_resource = Tasks(BASE_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/tasks_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Tasks/{1}".format(BASE_URI, TASK_SID)
        list_resource = Tasks(BASE_URI, AUTH)
        workflow = Task(list_resource, TASK_SID)
        workflow.update(attributes='attributes', workflow_sid='WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        exp_params = {
            'Attributes': "attributes",
            'WorkflowSid': "WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/tasks_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Tasks/{1}".format(BASE_URI, TASK_SID)
        list_resource = Tasks(BASE_URI, AUTH)
        list_resource.update(TASK_SID, attributes='attributes', workflow_sid='WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        exp_params = {
            'Attributes': "attributes",
            'WorkflowSid': "WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)



import unittest

from mock import patch

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.reservations import Reservations, Reservation


AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/Tasks/WTaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
BASE_WORKER_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/Workers/WKaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
RESERVATION_SID = "WRaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class TaskQueueTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/reservations_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Reservations/{1}".format(BASE_URI, RESERVATION_SID)
        list_resource = Reservations(BASE_URI, AUTH)
        list_resource.get(RESERVATION_SID)
        request.assert_called_with("GET", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/reservations_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Reservations".format(BASE_URI)
        list_resource = Reservations(BASE_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list_for_worker(self, request):
        resp = create_mock_json('tests/resources/task_router/reservations_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Reservations".format(BASE_WORKER_URI)
        list_resource = Reservations(BASE_WORKER_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/reservations_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Reservations/{1}".format(BASE_URI, RESERVATION_SID)
        list_resource = Reservations(BASE_URI, AUTH)
        workflow = Reservation(list_resource, RESERVATION_SID)
        workflow.update(worker_activity_sid='WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa', reservation_status='rejected')
        exp_params = {
            'WorkerActivitySid': "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            'ReservationStatus': "rejected"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/reservations_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Reservations/{1}".format(BASE_URI, RESERVATION_SID)
        list_resource = Reservations(BASE_URI, AUTH)
        list_resource.update(RESERVATION_SID, worker_activity_sid='WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
                             reservation_status='rejected')
        exp_params = {
            'WorkerActivitySid': "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            'ReservationStatus': "rejected"
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)



from twilio.rest.base import TwilioClient
from twilio.rest.resources import UNSET_TIMEOUT
from twilio.rest.resources.lookups.phone_numbers import PhoneNumbers


class TwilioLookupsClient(TwilioClient):
    """
    A client for accessing the Twilio Lookups API.

    The Twilio Lookups API provides information about phone numbers,
    including non-Twilio numbers. For more information, see the
    `Lookups API documentation <https://www.twilio.com/docs/XXX>`_.

    :param str account: Your Account Sid from `your dashboard
        <https://www.twilio.com/user/account>`_
    :param str token: Your Auth Token from `your dashboard
        <https://www.twilio.com/user/account>`_
    :param float timeout: The socket and read timeout for requests to Twilio
    """

    def __init__(self, account=None, token=None,
                 base="https://lookups.twilio.com", version="v1",
                 timeout=UNSET_TIMEOUT):

        super(TwilioLookupsClient, self).__init__(account, token, base,
                                                  version, timeout)

        self.version_uri = "%s/%s" % (base, version)
        self.phone_numbers = PhoneNumbers(self.version_uri, self.auth, timeout)



from . import transform_params
from . import InstanceResource, ListResource


class CallerId(InstanceResource):

    def delete(self):
        """
        Deletes this caller ID from the account.
        """
        return self.delete_instance()

    def update(self, **kwargs):
        """
        Update the CallerId
        """
        self.update_instance(**kwargs)


class CallerIds(ListResource):
    """ A list of :class:`CallerId` resources """

    name = "OutgoingCallerIds"
    key = "outgoing_caller_ids"
    instance = CallerId

    def delete(self, sid):
        """
        Deletes a specific :class:`CallerId` from the account.
        """
        self.delete_instance(sid)

    def list(self, **kwargs):
        """
        :param phone_number: Show caller ids with this phone number.
        :param friendly_name: Show caller ids with this friendly name.
        """
        return self.get_instances(kwargs)

    def update(self, sid, **kwargs):
        """
        Update a specific :class:`CallerId`
        """
        return self.update_instance(sid, kwargs)

    def validate(self, phone_number, **kwargs):
        """
        Begin the validation process for the given number.

        Returns a dictionary with the following keys

        **account_sid**:
        The unique id of the Account to which the Validation Request belongs.

        **phone_number**: The incoming phone number being validated,
        formatted with a '+' and country code e.g., +16175551212

        **friendly_name**: The friendly name you provided, if any.

        **validation_code**: The 6 digit validation code that must be entered
        via the phone to validate this phone number for Caller ID.

        :param phone_number: The phone number to call and validate
        :param friendly_name: A description for the new caller ID
        :param call_delay: Number of seconds to delay the validation call.
        :param extension: Digits to dial after connecting the validation call.
        :returns: A response dictionary
        """
        kwargs["phone_number"] = phone_number
        params = transform_params(kwargs)
        resp, validation = self.request("POST", self.uri, data=params)
        return validation



from mock import patch

from tests.tools import create_mock_json
from twilio.rest.monitor import TwilioMonitorClient


@patch("twilio.rest.resources.base.make_twilio_request")
def test_events(mock):
    client = TwilioMonitorClient("ACCOUNT_SID", "AUTH_TOKEN")
    resp = create_mock_json("tests/resources/monitor/events_instance.json")
    mock.return_value = resp
    client.events.get("AEaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
    uri = "https://monitor.twilio.com/v1/Events/AEaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    mock.assert_called_with("GET", uri, auth=("ACCOUNT_SID", "AUTH_TOKEN"), use_json_extension=False)



from mock import patch
from tests.tools import create_mock_json

from twilio.rest.resources.task_router.workers import Workers, Worker
from twilio.rest.resources.task_router.task_queues import TaskQueues, TaskQueue
from twilio.rest.resources.task_router.workflows import Workflows, Workflow

AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
WORKER_SID = "WKaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
TIMEOUT = 30


@patch("twilio.rest.resources.base.make_twilio_request")
def test_fetch_worker_statistics(request):
    resp = create_mock_json('tests/resources/task_router/workers_statistics_instance.json')
    resp.status_code = 200
    request.return_value = resp

    workers = Workers(BASE_URI, AUTH, TIMEOUT)
    worker = Worker(workers, 'WK123')
    worker.load_subresources()
    worker.statistics.get()
    request.assert_called_with(
        'GET',
        '{0}/Workers/WK123/Statistics'.format(BASE_URI),
        params={},
        auth=AUTH,
        timeout=TIMEOUT,
        use_json_extension=False,
    )


@patch("twilio.rest.resources.base.make_twilio_request")
def test_fetch_workers_statistics(request):
    resp = create_mock_json('tests/resources/task_router/workers_statistics_instance.json')
    resp.status_code = 200
    request.return_value = resp

    workers = Workers(BASE_URI, AUTH, TIMEOUT)
    workers.statistics.get()
    request.assert_called_with('GET',
                               '{0}/Workers/Statistics'.format(BASE_URI),
                               params={}, auth=AUTH, timeout=TIMEOUT,
                               use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_fetch_task_queue_statistics(request):
    resp = create_mock_json('tests/resources/task_router/task_queues_instance.json')
    resp.status_code = 200
    request.return_value = resp

    tqs = TaskQueues(BASE_URI, AUTH, 30)
    tq = TaskQueue(tqs, 'TQ123')
    tq.load_subresources()
    tq.statistics.get()
    request.assert_called_with('GET',
                               '{0}/TaskQueues/TQ123/Statistics'.format(BASE_URI),
                               params={}, auth=AUTH, timeout=30,
                               use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_fetch_task_queues_statistics(request):
    resp = create_mock_json('tests/resources/task_router/task_queues_instance.json')
    resp.status_code = 200
    request.return_value = resp

    tqs = TaskQueues(BASE_URI, AUTH, 30)
    tqs.statistics.get()
    request.assert_called_with('GET',
                               '{0}/TaskQueues/Statistics'.format(BASE_URI),
                               params={}, auth=AUTH, use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_fetch_task_workflow_statistics(request):
    resp = create_mock_json('tests/resources/task_router/workflows_instance.json')
    resp.status_code = 200
    request.return_value = resp

    workflows = Workflows(BASE_URI, AUTH)
    workflow = Workflow(workflows, 'WF123')
    workflow.load_subresources()
    workflow.statistics.get()
    request.assert_called_with('GET',
                               '{0}/Workflows/WF123/Statistics'.format(BASE_URI),
                               params={}, auth=AUTH, use_json_extension=False)



from datetime import date
import unittest

from mock import Mock
from nose.tools import assert_equal

from twilio.rest.resources import MediaList

DEFAULT = {
    'DateCreated<': None,
    'DateCreated>': None,
    'DateCreated': None,
}


class MediaTest(unittest.TestCase):

    def setUp(self):
        self.resource = MediaList("foo", ("sid", "token"))
        self.params = DEFAULT.copy()

    def test_list_on(self):
        self.resource.get_instances = Mock()
        self.resource.list(date_created=date(2011, 1, 1))
        self.params['DateCreated'] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_after(self):
        self.resource.get_instances = Mock()
        self.resource.list(after=date(2011, 1, 1))
        self.params['DateCreated>'] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_before(self):
        self.resource.get_instances = Mock()
        self.resource.list(before=date(2011, 1, 1))
        self.params['DateCreated<'] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_call(self):
        base_uri = self.resource.base_uri
        message_media = self.resource('MM123')
        assert_equal(message_media.base_uri, "%s/Messages/%s" % (base_uri, 'MM123'))



import unittest

from mock import patch

from tests.tools import create_mock_json
from twilio.rest.resources.conversations.conversations import ConversationsRoot


AUTH = ("AC123", "token")
BASE_URI = "https://conversations.twilio.com/v1"
CONVERSATION_SID = "CVaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class ConversationTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/conversations/conversation_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Conversations/{1}".format(BASE_URI, CONVERSATION_SID)
        list_resource = ConversationsRoot(BASE_URI, AUTH)
        list_resource.get(CONVERSATION_SID)
        request.assert_called_with("GET", uri, use_json_extension=False, auth=AUTH)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list_in_progress(self, request):
        resp = create_mock_json('tests/resources/conversations/conversation_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Conversations/InProgress".format(BASE_URI)
        list_resource = ConversationsRoot(BASE_URI, AUTH)
        list_resource.in_progress.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list_completed(self, request):
        resp = create_mock_json('tests/resources/conversations/conversation_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Conversations/Completed".format(BASE_URI)
        list_resource = ConversationsRoot(BASE_URI, AUTH)
        list_resource.completed.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH, use_json_extension=False)



from twilio.rest.resources.base import InstanceResource, ListResource


class SigningKey(InstanceResource):
    """
    A signing key resource.
    See https://www.twilio.com/docs/api/rest/signing-keys

    .. attribute:: sid

        The unique ID for this signing key.

    .. attribute:: friendly_name

        A human-readable description of this signing key.

    .. attribute:: secret

        This signing key's secret.

    .. attribute:: date_created

        The date this signing key was created, given as UTC in ISO 8601 format.

    .. attribute:: date_updated

        The date this singing key was last updated, given as UTC in ISO 8601
        format.
    """

    def update(self, **kwargs):
        """
        Update this signing key
        """
        return self.parent.update(self.name, **kwargs)

    def delete(self):
        """
        Delete this signing key
        """
        return self.parent.delete(self.name)


class SigningKeys(ListResource):
    name = "SigningKeys"
    key = "signing_keys"
    instance = SigningKey

    def create(self, **kwargs):
        """
        Create a :class:`SigningKey` with any of these optional parameters.

        :param friendly_name: A human readable description of the signing key.
        """
        return self.create_instance(kwargs)

    def update(self, sid, **kwargs):
        """
        Update a :class:`SigningKey` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)

    def delete(self, sid):
        """
        Delete a :class:`SigningKey`
        """
        return self.delete_instance(sid)

    def list(self, **kw):
        """
        List is not supported, hence raises an Error
        """
        raise AttributeError("SigningKeys do not support lists()")



from .. import NextGenInstanceResource, NextGenListResource
from .statistics import Statistics
from .reservations import Reservations


class Worker(NextGenInstanceResource):
    """
    A Worker resource.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/workers>_`
    for more information.

    .. attribute:: sid

        The unique ID of this Worker.

    .. attribute:: account_sid

        The unique ID of the account that owns this worker.

    .. attribute:: workspace_sid

        The unique ID of the :class:`Workspace` that contains this worker.

    .. attribute:: friendly_name

        A human-readable name for this worker.

    .. attribute:: attributes

        A JSON object describing this worker. For example, for a Worker that
        handles English language phone calls:
            '{"language":"english","task-type":"phone"}'
        These attributes determine which :class:`TaskQueue` this worker will
        subscribe to. These attributes will also be passed to the Assignment
        Callback URL whenever TaskRouter assigns a :class:`Task` to this
        worker, so you can also use this as a place to store information that
        you'll need when routing a Task to the Worker (for example, the
        Worker's phone number or Twilio Client name).

    .. attribute:: available

        A Boolean value indicating whether the worker can be assigned another
        task. When true, the worker can be assigned a new task; when false,
        the worker will not be assigned any tasks.

    .. attribute:: activity_sid

        The unique ID of the :class:`Activity` this Worker is currently
        performing.

    .. attribute:: activity_name

        A string describing the worker's current activity. Workers may only
        perform Activities that exist in this Workspace.

    .. attribute:: date_created

        The time this worker was created, given in UTC ISO 8601 format.

    .. attribute:: date_updated

        The time this worker was last updated, given in UTC ISO 8601 format.

    .. attribute:: date_status_changed

        The time of the last change to this worker's activity. Used to
        calculate :class: `Workflow` statistics.
    """
    subresources = [
        Statistics,
        Reservations
    ]

    def delete(self):
        """
        Delete a worker.
        """
        return self.parent.delete_instance(self.name)

    def update(self, **kwargs):
        """
        Update a worker.
        """
        return self.parent.update_instance(self.name, kwargs)


class Workers(NextGenListResource):
    """ A list of Worker resources """

    name = "Workers"
    instance = Worker

    def __init__(self, base_uri, auth, timeout, **kwargs):
        super(Workers, self).__init__(base_uri, auth, timeout, **kwargs)
        self.statistics = Statistics(self.uri, auth, timeout, **kwargs)

    def create(self, friendly_name, **kwargs):
        """
        Create a Workflow.

        :param friendly_name: String representing user-friendly name for the
            Worker.
        :param activity_sid: A valid Activity describing the worker's initial
            state.
        :param attributes: JSON object describing this worker. For example:
            { 'email: 'Bob@foo.com', 'phone': '8675309' }. This data will be
            passed to the Assignment Callback URL whenever Work Distribution
            Service assigns a Task to this worker.
        """
        kwargs['friendly_name'] = friendly_name
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete the given worker
        """
        return self.delete_instance(sid)

    def update(self, sid, **kwargs):
        """
        Update a :class:`Worker` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)



import unittest
from nose.tools import assert_equal

from mock import patch
from tests.tools import create_mock_json

from twilio.rest.resources.pricing.voice import (
    VoiceCountries,
    VoiceNumbers,
)


AUTH = ("AC123", "token")
BASE_URI = "https://pricing.twilio.com/v1"


class VoiceTest(unittest.TestCase):

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_voice_countries(self, request):
        resp = create_mock_json('tests/resources/pricing/voice_countries_list.json')
        resp.status_code = 200
        request.return_value = resp

        countries = VoiceCountries(BASE_URI + "/Voice", AUTH)
        result = countries.list()

        assert_equal(result[0].iso_country, "AC")
        assert_equal(len(result), 3)

        request.assert_called_with(
            "GET",
            "{0}/Voice/Countries".format(BASE_URI),
            auth=AUTH,
            use_json_extension=False,
        )

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_voice_country(self, request):
        resp = create_mock_json('tests/resources/pricing/voice_country_instance.json')
        resp.status_code = 200
        request.return_value = resp

        countries = VoiceCountries(BASE_URI + "/Voice", AUTH)
        country = countries.get('EE')

        assert_equal(country.country, "Estonia")
        assert_equal(
            country.inbound_call_prices[0],
            {
                'number_type': 'mobile',
                'call_base_price': 0.0075,
                'call_current_price': 0.0070
            },
        )

        request.assert_called_with(
            "GET",
            "{0}/Voice/Countries/EE".format(BASE_URI),
            auth=AUTH,
            use_json_extension=False,
        )

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_voice_number(self, request):
        resp = create_mock_json('tests/resources/pricing/voice_number_instance.json')
        resp.status_code = 200
        request.return_value = resp

        numbers = VoiceNumbers(BASE_URI + "/Voice", AUTH)
        result = numbers.get('+14089673429')

        assert_equal(result.number, '+14089673429')
        assert_equal(result.price_unit, 'usd')

        request.assert_called_with(
            "GET",
            "{0}/Voice/Numbers/+14089673429".format(BASE_URI),
            auth=AUTH,
            use_json_extension=False,
        )



import unittest

from mock import patch

from twilio.rest.resources import Tokens


DEFAULT = {
    'Ttl': None,
}


class TokensTest(unittest.TestCase):

    def setUp(self):
        self.resource = Tokens('foo', ('sid', 'token'))
        self.params = DEFAULT.copy()

    def test_create(self):
        with patch.object(self.resource, 'create_instance') as mock:
            self.resource.create(
                ttl=111,
            )
            mock.assert_called_with(
                {
                    'ttl': 111,
                }
            )
